import os
import random
import secrets
import sys
from functools import partial
from typing import Callable, Literal, Optional, Union

import torch
from torch import distributed as dist

from lhotse.utils import fix_random_seed

LHOTSE_PROCESS_SEED = "LHOTSE_PROCESS_SEED"


def make_worker_init_fn(
    rank: Optional[int] = None,
    world_size: Optional[int] = None,
    set_different_node_and_worker_seeds: bool = True,
    seed: Optional[int] = 42,
) -> Optional[Callable[[int], None]]:
    """
    Calling this function creates a worker_init_fn suitable to pass to PyTorch's DataLoader.

    It helps with two issues:

    * sets the random seeds differently for each worker and node, which helps with
        avoiding duplication in randomized data augmentation techniques.
    * sets environment variables that help WebDataset detect it's inside multi-GPU (DDP)
        training, so that it correctly de-duplicates the data across nodes.
    """
    return partial(
        worker_init_fn,
        rank=rank,
        world_size=world_size,
        set_different_node_and_worker_seeds=set_different_node_and_worker_seeds,
        seed=seed,
    )


def worker_init_fn(
    worker_id: int,
    rank: Optional[int] = None,
    world_size: Optional[int] = None,
    set_different_node_and_worker_seeds: bool = True,
    seed: Optional[int] = 42,
) -> None:
    """
    Function created by :func:`~lhotse.dataset.dataloading.make_worker_init_fn`, refer to its documentation for details.
    """
    if set_different_node_and_worker_seeds:
        process_seed = seed + 100 * worker_id
        if rank is not None:
            process_seed += 100000 * rank
        fix_random_seed(process_seed)
        os.environ[LHOTSE_PROCESS_SEED] = str(process_seed)

    if rank is None and world_size is None:
        return

    assert (
        rank is not None and world_size is not None
    ), f"Both args must be not None: rank={rank}, world_size={world_size}"

    # This sets the rank/world_size info for WebDataset to read it in worker subprocesses.
    # If we didn't do it, WebDataset will "think" this is always single-node training,
    # because DataLoader workers did not initialize torch.distributed.
    os.environ["RANK"] = str(rank)
    os.environ["WORLD_SIZE"] = str(world_size)


def resolve_seed(seed: Union[int, Literal["trng", "randomized"], None]) -> int:
    """
    Resolves the special values of random seed supported in Lhotse.

    If it's an integer, we'll just return it.

    If it's "trng", we'll use the ``secrets`` module to generate a random seed
    using a true RNG (to the extend supported by the OS).

    If it's "randomized", we'll check whether we're in a dataloading worker of ``torch.utils.data.DataLoader``.
    If we are, we expect that it was passed the result of :func:`~lhotse.dataset.dataloading.make_worker_init_fn`
    into its ``worker_init_fn`` argument, in which case we'll return a special seed exclusive to that worker.
    If we are not in a dataloading worker (or ``num_workers`` was set to ``0``), we'll return Python's ``random``
    module global seed.
    """

    # Specific number provided: use it.
    if isinstance(seed, int):
        return seed

    # No request for a specific type of random seed resolution: return Python's global random seed.
    if seed is None:
        return random.getstate()[1][0]

    # Deterministic randomized random seed resolution:
    # Each dataloading worker and DDP rank gets a separate random seed.
    # If we're not in a dataloading worker, use global RNG's current seed.
    if seed == "randomized":
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            # Not in a dataloader sub-process: get Python's global random seed.
            return random.getstate()[1][0]
        else:
            # In a dataloader sub-process: read out the seed we assigned to it.
            assert LHOTSE_PROCESS_SEED in os.environ, (
                "Requested seed='randomized' for shuffling shards differently "
                "on each DataLoader node and worker, "
                "but lhotse.dataset.dataloading.worker_init_fn was not called."
            )
            return int(os.environ[LHOTSE_PROCESS_SEED])

    # True-random number generator requested for seed generation ("complete randomness").
    if seed == "trng":
        # 2**32 may trigger the following exception if you add anything:
        # File "_mt19937.pyx", line 180, in numpy.random._mt19937.MT19937._legacy_seeding
        # ValueError: Seed must be between 0 and 2**32 - 1
        return secrets.randbelow(2**31)

    raise ValueError(
        f"Unexpected type or value of seed: {type(seed)=} {seed=}. "
        f"Supported values are: None, int, 'trng', and 'randomized'."
    )


def get_world_size() -> int:
    """Source: https://github.com/danpovey/icefall/blob/74bf02bba6016c1eb37858a4e0e8a40f7d302bdb/icefall/dist.py#L56"""
    if "WORLD_SIZE" in os.environ:
        return int(os.environ["WORLD_SIZE"])
    if dist.is_available() and dist.is_initialized():
        return dist.get_world_size()
    else:
        return 1


def get_rank() -> int:
    """Source: https://github.com/danpovey/icefall/blob/74bf02bba6016c1eb37858a4e0e8a40f7d302bdb/icefall/dist.py#L56"""
    if "RANK" in os.environ:
        return int(os.environ["RANK"])
    elif dist.is_available() and dist.is_initialized():
        return dist.get_rank()
    else:
        return 0



import tarfile
from pathlib import Path
from typing import Generator, Optional, Tuple, Union

from lhotse import Features, Recording
from lhotse.array import Array, TemporalArray
from lhotse.serialization import decode_json_line, deserialize_item, open_best
from lhotse.shar.utils import fill_shar_placeholder
from lhotse.utils import Pathlike

Manifest = Union[Recording, Array, TemporalArray, Features]


class TarIterator:
    """
    TarIterator is a convenience class for reading arrays/audio stored in Lhotse Shar tar files.
    It is specific to Lhotse Shar format and expects the tar file to have the following structure:

    * each file is stored in a separate tar member
    * the file name is the key of the array
    * every array has two corresponding files:
        * the metadata: the file extension is ``.json`` and the file contains
          a Lhotse manifest (Recording, Array, TemporalArray, Features)
          for the data item.
        * the data: the file extension is the format of the array,
          and the file contents are the serialized array (possibly compressed).
        * the data file can be empty in case some cut did not contain that field.
          In that case, the data file has extension ``.nodata`` and the manifest file
          has extension ``.nometa``.
        * these files are saved one after another, the data is first, and the metadata follows.

    Iterating over TarReader yields tuples of ``(Optional[manifest], filename)`` where
    ``manifest`` is a Lhotse manifest with binary data attached to it, and ``filename``
    is the name of the data file inside tar archive.
    """

    def __init__(self, source: Pathlike) -> None:
        self.source = source

    def __iter__(
        self,
    ) -> Generator[Tuple[Optional[Manifest], Path], None, None]:
        with tarfile.open(fileobj=open_best(self.source, mode="rb"), mode="r|*") as tar:
            for ((data, data_path), (meta, meta_path)) in iterate_tarfile_pairwise(tar):
                if meta is not None:
                    meta = deserialize_item(decode_json_line(meta.decode("utf-8")))
                    fill_shar_placeholder(manifest=meta, data=data, tarpath=data_path)
                yield meta, data_path


def iterate_tarfile_pairwise(
    tar_file: tarfile.TarFile,
) -> Generator[Tuple[Optional[bytes], Optional[Manifest], Path, Path], None, None]:
    result = []
    for tarinfo in tar_file:
        if len(result) == 2:
            yield tuple(result)
            result = []
        result.append(parse_tarinfo(tarinfo, tar_file))

    if len(result) == 2:
        yield tuple(result)

    if len(result) == 1:
        raise RuntimeError(
            "Uneven number of files in the tarfile (expected to iterate pairs of binary data + JSON metadata."
        )


def parse_tarinfo(
    tarinfo: tarfile.TarInfo, tar_file: tarfile.TarFile
) -> Tuple[Optional[bytes], Path]:
    """
    Parse a tarinfo object and return the data it points to as well as the internal path.
    """
    path = Path(tarinfo.path)
    if path.suffix == ".nodata" or path.suffix == ".nometa":
        return None, path
    data = tar_file.extractfile(tarinfo).read()
    return data, path



"""
This recipe provides functionality for downloading and preparing the fleurs
corpus. The data is hosted on huggingface and to enable more control of the
download format, we use the streaming download interface and save each audio
file as it is streamed. The download can take quite some time.

The fleurs corpus consist of data in 102 languages spoken by multiple speakers.
There is about 10 hrs of trainign data in each language with smaller
accompanying dev and test sets. Full details can be found in

@inproceedings{conneau2023fleurs,
  title={Fleurs: Few-shot learning evaluation of universal representations of speech},
  author={Conneau, Alexis and Ma, Min and Khanuja, Simran and Zhang, Yu and Axelrod, Vera and Dalmia, Siddharth and Riesa, Jason and Rivera, Clara and Bapna, Ankur},
  booktitle={2022 IEEE Spoken Language Technology Workshop (SLT)},
  pages={798--805},
  year={2023},
  organization={IEEE}
}
"""
import logging
from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from tqdm import tqdm

from lhotse import (
    Recording,
    RecordingSet,
    SupervisionSegment,
    SupervisionSet,
    audio,
    fix_manifests,
    get_ffmpeg_torchaudio_info_enabled,
    set_ffmpeg_torchaudio_info_enabled,
)
from lhotse.parallel import parallel_map
from lhotse.utils import Pathlike, is_module_available

# The FLEURS languages are indicated by 2-letter ISO-codes followed by a
# country code, i.e.,
#
#  en_us, fr_fr, ml_in
#
# for American English, French French and Indian Malayalam respectively.

DEFAULT_LANGUAGES = [
    "af_za",
    "am_et",
    "ar_eg",
    "as_in",
    "ast_es",
    "az_az",
    "be_by",
    "bg_bg",
    "bn_in",
    "bs_ba",
    "ca_es",
    "ceb_ph",
    "ckb_iq",
    "cmn_hans_cn",
    "cs_cz",
    "cy_gb",
    "da_dk",
    "de_de",
    "el_gr",
    "en_us",
    "es_419",
    "et_ee",
    "fa_ir",
    "ff_sn",
    "fi_fi",
    "fil_ph",
    "fr_fr",
    "ga_ie",
    "gl_es",
    "gu_in",
    "ha_ng",
    "he_il",
    "hi_in",
    "hr_hr",
    "hu_hu",
    "hy_am",
    "id_id",
    "ig_ng",
    "is_is",
    "it_it",
    "ja_jp",
    "jv_id",
    "ka_ge",
    "kam_ke",
    "kea_cv",
    "kk_kz",
    "km_kh",
    "kn_in",
    "ko_kr",
    "ky_kg",
    "lb_lu",
    "lg_ug",
    "ln_cd",
    "lo_la",
    "lt_lt",
    "luo_ke",
    "lv_lv",
    "mi_nz",
    "mk_mk",
    "ml_in",
    "mn_mn",
    "mr_in",
    "ms_my",
    "mt_mt",
    "my_mm",
    "nb_no",
    "ne_np",
    "nl_nl",
    "nso_za",
    "ny_mw",
    "oc_fr",
    "om_et",
    "or_in",
    "pa_in",
    "pl_pl",
    "ps_af",
    "pt_br",
    "ro_ro",
    "ru_ru",
    "sd_in",
    "sk_sk",
    "sl_si",
    "sn_zw",
    "so_so",
    "sr_rs",
    "sv_se",
    "sw_ke",
    "ta_in",
    "te_in",
    "tg_tj",
    "th_th",
    "tr_tr",
    "uk_ua",
    "umb_ao",
    "ur_pk",
    "uz_uz",
    "vi_vn",
    "wo_sn",
    "xh_za",
    "yo_ng",
    "yue_hant_hk",
    "zu_za",
]


def download_fleurs(
    target_dir: Pathlike = ".",
    languages: Optional[Union[str, Sequence[str]]] = "all",
    force_download: Optional[bool] = False,
) -> Path:
    """
    Download the specified fleurs datasets.

    :param target_dir: The path to which the corpus will be downloaded.
    :type target_dir: Pathlike
    :param languages: Optional list of str or str specifying which
        languages to download. The str specifier for a language has the
        ISOCODE_COUNTRYCODE format, and is all lower case. By default
        this is set to "all", which will download the entire set of
        languages.
    :type languages: Optional[Union[str, Sequence[str]]]
    :param force_download: Specifies whether to overwrite an existing
        archive.
    :type force_download: bool
    :return: The root path of the downloaded data
    :rtype: Path
    """
    target_dir = Path(target_dir)
    corpus_dir = target_dir / "fleurs"
    metadata_dir = corpus_dir / "metadata"
    metadata_dir.mkdir(parents=True, exist_ok=True)

    if isinstance(languages, str) and languages == "all" or languages[0] == "all":
        languages = DEFAULT_LANGUAGES

    if isinstance(languages, str):
        languages = [languages]

    for lang in tqdm(languages):
        # Download one language at a time
        lang_dir = corpus_dir / lang
        download_single_fleurs_language(
            lang_dir,
            lang,
            force_download,
        )
    return corpus_dir


def download_single_fleurs_language(
    target_dir: Pathlike,
    language: str,
    force_download: bool = False,
) -> Path:
    """
    Download a single fleurs language

    :param target_dir: The path to which one langauge will be downloaded
    :type target_dir: Pathlike
    :param language: The code for the specified language
    :type language: str
    :param force_download: Specifies whether to overwrite an existing
        archive.
    :type force_download: bool
    :return: The path to the downloaded data for the specified language
    :rtype: Path
    """
    if not is_module_available("datasets"):
        raise ImportError(
            "The huggingface datasets package is not installed. Please install"
            " ...(pip install datasets)"
        )
    else:
        from datasets import load_dataset

    def _identity(x):
        return x

    target_dir = Path(target_dir)
    metadata_dir = target_dir.parents[0] / "metadata" / language
    target_dir.mkdir(parents=True, exist_ok=True)
    metadata_dir.mkdir(parents=True, exist_ok=True)

    completed_detector = target_dir / f".{language}_completed"
    if completed_detector.is_file() and not force_download:
        logging.info("Skipping dowload because {completed_detector} exists.")
        return target_dir

    for split in tqdm(["train", "validation", "test"]):
        fleurs = load_dataset(
            "google/fleurs",
            language,
            cache_dir=target_dir,
            streaming=True,
            split=split,
        )
        metadata = []
        osplit = "dev" if split == "validation" else split
        split_dir = target_dir / osplit
        split_dir.mkdir(parents=True, exist_ok=True)
        for data in tqdm(fleurs, desc=f"Downloading data from {language}-{osplit}"):
            audio.save_audio(
                f"{split_dir}/{Path(data['audio']['path']).name}",
                data["audio"]["array"],
                data["audio"]["sampling_rate"],
            )
            metadata_ = [
                str(data["id"]),  # ID
                Path(data["audio"]["path"]).name,  # filename
                data["raw_transcription"],  # raw transcript
                data["transcription"],  # transcript
                " ".join("|".join(data["transcription"].split())) + " |",  # chars
                str(data["num_samples"]),  # number of audio samples
                "FEMALE" if data["gender"] == 1 else "MALE",  # gender
            ]
            metadata.append(metadata_)
        with open(metadata_dir / f"{osplit}.tsv", "w") as f:
            for md in metadata:
                print("\t".join(md), file=f)

    completed_detector.touch()
    return target_dir


def prepare_fleurs(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    languages: Optional[Union[str, Sequence[str]]] = "all",
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]]:
    """
    Prepares the manifest for all of the FLEURS languages requested.

    :param corpus_dir: Path to the root where the FLEURS data are stored.
    :type corpus_dir: Pathlike,
    :param output_dir: The directory where the .jsonl.gz manifests will be written.
    :type output_dir: Pathlike,
    :param langauges: str or str sequence specifying the languages to prepare.
        The str 'all' prepares all 102 languages.
    :return: The manifest
    :rtype: Dict[str, Dict[str, Union[RecordingSet, Supervisions]]]]
    """

    if isinstance(corpus_dir, str):
        corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    if isinstance(output_dir, str):
        output_dir = Path(output_dir)

    output_dir.mkdir(mode=511, parents=True, exist_ok=True)

    langs_list = DEFAULT_LANGUAGES
    if isinstance(languages, str) and languages != "all":
        langs_list = [languages]
    elif isinstance(languages, list) or isinstance(languages, tuple):
        if languages[0] != "all":
            langs_list = languages

    # Start buildings the recordings and supervisions
    manifests = {}
    for lang in langs_list:
        corpus_dir_lang = corpus_dir / f"{lang}"
        if not corpus_dir_lang.is_dir():
            logging.info(f"Skipping {lang}. No directory {corpus_dir_lang} found.")
            continue
        output_dir_lang = output_dir / f"{lang}"
        output_dir_lang.mkdir(mode=511, parents=True, exist_ok=True)
        manifests[lang] = prepare_single_fleurs_language(
            corpus_dir_lang,
            output_dir_lang,
            language=lang,
            num_jobs=num_jobs,
        )

    if output_dir is not None:
        for l in manifests:
            for dset in ("train", "dev", "test"):
                manifests[l][dset]["supervisions"].to_file(
                    output_dir / f"{l}" / f"fleurs-{l}_supervisions_{dset}.jsonl.gz"
                )
                manifests[l][dset]["recordings"].to_file(
                    output_dir / f"{l}" / f"fleurs-{l}_recordings_{dset}.jsonl.gz"
                )
    return manifests


def _make_recording(path):
    return Recording.from_file(path, recording_id=Path(path).stem)


def prepare_single_fleurs_language(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    language: str = "language",
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Prepares manifests using a single FLEURS language.

    :param corpus_dir: Path to the root where the FLEURS data are stored.
    :type corpus_dir: Pathlike,
    :param output_dir: The directory where the .jsonl.gz manifests will be written.
    :type output_dir: Pathlike,
    :param langauge: str specifying the language to prepare.

    :return: The manifest
    :rtype: Dict[str, Dict[str, Union[RecordingSet, Supervisions]]]]
    """

    if isinstance(corpus_dir, str):
        corpus_dir = Path(corpus_dir)

    recordings = {"train": [], "dev": [], "test": []}
    supervisions = {"train": [], "dev": [], "test": []}

    # First prepare the supervisions
    for dset in ("train", "dev", "test"):
        print(f"Preparing {dset} ...")
        prompt_ids = {}
        with open(
            corpus_dir.parents[0] / "metadata" / corpus_dir.stem / f"{dset}.tsv"
        ) as f:
            for l in f:
                vals = l.strip().split("\t")
                prompt_id, fname, raw_text, text, _, nsamples, gender = vals
                if prompt_id not in prompt_ids:
                    prompt_ids[prompt_id] = 0
                prompt_ids[prompt_id] += 1
                fname = Path(fname).stem
                supervisions[dset].append(
                    SupervisionSegment(
                        id=f"{prompt_id}_{prompt_ids[prompt_id]}_{fname}",
                        recording_id=fname,
                        start=0.0,
                        duration=round(int(nsamples) / 16000, 4),
                        channel=0,
                        text=text,
                        language=language,
                        speaker=f"{prompt_id}_{prompt_ids[prompt_id]}",
                        gender=gender,
                        custom={"raw_text": raw_text},
                    )
                )
    for dset in ("train", "dev", "test"):
        for reco in tqdm(
            parallel_map(
                _make_recording,
                (
                    corpus_dir / f"{dset}/{s.recording_id}.wav"
                    for s in supervisions[dset]
                ),
                num_jobs=num_jobs,
            ),
            desc=f"Making recordings from {language} {dset}",
        ):
            recordings[dset].append(reco)
    manifests = {}
    for dset in ("train", "dev", "test"):
        sups = SupervisionSet.from_segments(supervisions[dset])
        recos = RecordingSet.from_recordings(recordings[dset])
        recos, sups = fix_manifests(recos, sups)
        manifests[dset] = {"supervisions": sups, "recordings": recos}
    return manifests



"""
The Edinburgh International Accents of English Corpus

Citation

Sanabria, Ramon; Markl, Nina; Carmantini, Andrea; Klejch, Ondrej; Bell, Peter; Bogoychev, Nikolay. (2023). The Edinburgh International Accents of English Corpus, [dataset]. University of Edinburgh. School of Informatics. The Institute for Language, Cognition and Computation. The Centre for Speech Technology Research. https://doi.org/10.7488/ds/3832.

Description

English is the most widely spoken language in the world, used daily by millions of people as a first or second language in many different contexts.
As a result, there are many varieties of English.
Although the great many advances in English automatic speech recognition (ASR) over the past decades, results are usually reported based on test datasets which fail to represent the diversity of English as spoken today around the globe.
We present the first release of The Edinburgh International Accents of English Corpus (EdAcc).
This dataset attempts to better represent the wide diversity of English, encompassing almost 40 hours of dyadic video call conversations between friends.
Unlike other datasets, EdAcc includes a wide range of first and second-language varieties of English and a linguistic background profile of each speaker.
Results on latest public, and commercial models show that EdAcc highlights shortcomings of current English ASR models.
The best performing model, trained on 680 thousand hours of transcribed data, obtains an average of 19.7% WER -- in contrast to the the 2.7% WER obtained when evaluated on US English clean read speech.
Across all models, we observe a drop in performance on Jamaican, Indonesian, Nigerian, and Kenyan English speakers.
Recordings, linguistic backgrounds, data statement, and evaluation scripts are released on our website under CC-BY-SA.

Source: https://datashare.ed.ac.uk/handle/10283/4836
"""
import logging
import math
import shutil
import tarfile
import zipfile
from pathlib import Path
from typing import Dict, Optional, Union

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import RecordingSet
from lhotse.supervision import SupervisionSet
from lhotse.utils import Pathlike, is_module_available, resumable_download, safe_extract

_EDACC_SAMPLING_RATE = 32000


def download_edacc(
    target_dir: Pathlike = ".",
    force_download: bool = False,
    base_url: str = "https://datashare.ed.ac.uk/download/",
) -> Path:
    """
    Download and extract the EDACC dataset.

    :param target_dir: Pathlike, the path of the dir to store the dataset.
    :param force_download: Bool, if True, download the data even if it exists.
    :param base_url: str, the url of the website used to fetch the archive from.
    :return: the path to downloaded and extracted directory with data.
    """
    archive_name = "DS_10283_8983.zip"

    target_dir = Path(target_dir)
    corpus_dir = target_dir / "edacc"
    target_dir.mkdir(parents=True, exist_ok=True)

    completed_detector = corpus_dir / ".completed"
    if completed_detector.is_file():
        logging.info(f"Skipping EDACC download because {completed_detector} exists.")
        return corpus_dir

    # Maybe-download the archive.
    archive_path = target_dir / archive_name
    resumable_download(
        f"{base_url}/{archive_name}",
        filename=archive_path,
        force_download=force_download,
    )

    # Remove partial unpacked files, if any, and unpack everything.
    shutil.rmtree(corpus_dir, ignore_errors=True)
    with zipfile.ZipFile(archive_path) as zip:
        zip.extractall(path=corpus_dir)
    tar_name = "edacc_v1.0.tar.gz"
    with tarfile.open(corpus_dir / tar_name) as tar:
        safe_extract(tar, corpus_dir)
    completed_detector.touch()

    return corpus_dir


def prepare_edacc(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.

    :param corpus_dir: a path to the unzipped EDACC directory (has ``edacc_v1.0`` inside).
    :param output_dir: an optional path where to write the manifests.
    :return: a dict with structure ``{"dev|test": {"recordings|supervisions": <manifest>}}``
    """
    from lhotse.kaldi import load_kaldi_data_dir

    if not is_module_available("pandas"):
        raise ValueError("Please install pandas via 'pip install pandas'.")

    corpus_dir = Path(corpus_dir) / "edacc_v1.0"
    audio_dir = corpus_dir / "data"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
    manifests = {}

    # Read extra metadata
    spk2meta = parse_linguistic_background(corpus_dir / "linguistic_background.csv")

    # Create recordings manifest and prepare data to create wav.scp files later.
    recordings = RecordingSet.from_dir(audio_dir, "*.wav")
    for r in recordings:
        assert r.num_channels == 1, f"Unexpected multi-channel recording: {r}"
        assert r.sampling_rate == _EDACC_SAMPLING_RATE
    wav_scp = {r.id: f"{r.id} {r.sources[0].source}" for r in recordings}

    for split in ("dev", "test"):
        data_dir = corpus_dir / split

        # First, create wav.scp in Kaldi data dir, and then just import it.
        with open(data_dir / "segments") as f:
            split_rec_ids = set(l.split()[1] for l in f)
        with open(data_dir / "wav.scp", "w") as f:
            for rid, rstr in sorted(wav_scp.items()):
                if rid in split_rec_ids:
                    print(rstr, file=f)
        recordings, supervisions, _ = load_kaldi_data_dir(
            data_dir, sampling_rate=_EDACC_SAMPLING_RATE
        )

        # Add extra metadata available.
        with open(data_dir / "conv.list") as f:
            conv_rec_ids = set(map(str.strip, f))
        for s in supervisions:
            s.language = "English"
            s.is_conversational = s.recording_id in conv_rec_ids
            for key, val in spk2meta[s.speaker].items():
                setattr(s, key, val)

        # Fix, validate, and save manifests.
        recordings, supervisions = fix_manifests(recordings, supervisions)
        validate_recordings_and_supervisions(recordings, supervisions)
        manifests[split] = {"recordings": recordings, "supervisions": supervisions}
        if output_dir is not None:
            recordings.to_file(output_dir / f"edacc_recordings_{split}.jsonl.gz")
            supervisions.to_file(output_dir / f"edacc_supervisions_{split}.jsonl.gz")

    return manifests


def parse_linguistic_background(path: Pathlike) -> Dict:
    import pandas as pd

    df = pd.read_csv(path)
    df = df.rename(
        columns={
            "What is your gender?": "gender",
            "What’s your ethnic background? ": "ethnicity",
            "What is your higher level of education?": "education",
            "How would you describe your accent in English? (e.g. Italian, Glaswegian)": "accent",
            "Do you speak any second languages? separate them with commas  (e.g., Mandarin,Catalan,French )": "other_languages",
            "What’s your year of birth? (e.g., 1992)": "birth_year",
            "What year did you start learning English? (e.g., 1999)": "start_english_year",
        }
    )
    df["age"] = 2022 - df.birth_year
    df["years_speaking_english"] = 2022 - df.start_english_year

    def parse(key, val) -> Optional:
        if key == "years_speaking_english":
            if math.isnan(val):
                return None
            return int(val)
        if key == "other_languages":
            if isinstance(val, float) and math.isnan(val):
                return []
            return [v.strip() for v in val.split(",")]
        if isinstance(val, str):
            return val.strip()
        return val

    spk2meta = {
        row["PARTICIPANT_ID"]: {
            m: parse(m, row[m])
            for m in (
                "gender",
                "ethnicity",
                "education",
                "accent",
                "other_languages",
                "age",
                "years_speaking_english",
            )
        }
        for _, row in df.iterrows()
    }

    return spk2meta



"""
Note: this module is very heavily based on a torchaudio tutorial about forced
alignment with Wav2Vec2 created by Moto Hira.

Link: https://pytorch.org/audio/stable/pipelines.html
"""
from functools import partial
from typing import Generator

from lhotse import CutSet, MonoCut
from lhotse.parallel import ParallelExecutor

from .asr_aligner import ASRForcedAligner
from .mms_aligner import MMSForcedAligner


def __get_aligner_class(bundle_name: str):
    if bundle_name == "MMS_FA":
        return MMSForcedAligner
    elif "ASR" in bundle_name:
        return ASRForcedAligner
    else:
        raise ValueError(f"Unknown bundle name: {bundle_name}")


def align_with_torchaudio(
    cuts: CutSet,
    bundle_name: str = "WAV2VEC2_ASR_BASE_960H",
    device: str = "cpu",
    normalize_text: bool = True,
    num_jobs: int = 1,
    verbose: bool = False,
    check_language: bool = True,
) -> Generator[MonoCut, None, None]:
    """
    Use a pretrained model from torchaudio (such as Wav2Vec2) to perform forced
    word-level alignment of a CutSet.

    This means that for every SupervisionSegment with a transcript, we will find the
    start and end timestamps for each of the words.

    We support cuts with multiple supervisions -- the forced alignment will be done
    for every supervision region separately, and attached to the relevant supervision.

    Note: this is an experimental feature of Lhotse, and is not guaranteed to yield
    high quality of data.

    See torchaudio's documentation and tutorials for more details:
    - https://pytorch.org/audio/stable/tutorials/forced_alignment_tutorial.html
    - https://pytorch.org/audio/stable/pipelines.html

    :param cuts: input CutSet.
    :param bundle_name: name of the selected pretrained model from torchaudio.
        By default, we use WAV2VEC2_ASR_BASE_960H (English only).
        In order to use a multilingual alignment model, use bundle_name='MMS_FA'.
    :param device: device on which to run the computation.
    :param normalize_text: by default, we'll try to normalize the text by making
        it uppercase and discarding symbols outside of model's character level vocabulary.
        If this causes issues, turn the option off and normalize the text yourself.
    :param num_jobs: number of parallel jobs to use for alignment.
    :param verbose: if True, display the progress bar.
    :param check_language: if False, disable warning about the non-existent language tags in supervisions.
    :return: a generator of cuts that have the "alignment" field set in each of
        their supervisions.
    """
    AlignerClass = __get_aligner_class(bundle_name)
    aligner_init_fn = partial(
        AlignerClass,
        bundle_name=bundle_name,
        device=device,
        check_language=check_language,
    )
    processor = ParallelExecutor(
        init_fn=aligner_init_fn,
        num_jobs=num_jobs,
        verbose=verbose,
        description="Aligning",
    )
    return processor(cuts, normalize=normalize_text)



"""
The following are the original TED-LIUM 3 README contents.

This is the TED-LIUM corpus release 3,
licensed under Creative Commons BY-NC-ND 3.0 (http://creativecommons.org/licenses/by-nc-nd/3.0/deed.en).

All talks and text are property of TED Conferences LLC.

This new TED-LIUM release was made through a collaboration between the Ubiqus company and the LIUM (University of Le Mans, France)

---

Contents:

- 2351 audio talks in NIST sphere format (SPH), including talks from TED-LIUM 2: be careful, same talks but not same audio files (only these audio file must be used with the TED-LIUM 3 STM files)
--> 452 hours of audio
- 2351 aligned automatic transcripts in STM format

- TEDLIUM 2 dev and test data: 19 TED talks in SPH format with corresponding manual transcriptions (cf. 'legacy' distribution below).

- Dictionary with pronunciations (159848 entries), same file as the one included in TED-LIUM 2
- Selected monolingual data for language modeling from WMT12 publicly available corpora: these files come from the TED-LIUM 2 release, but have been modified to get a tokenization more relevant for English language

- Two corpus distributions:
-- the legacy one, on which the dev and test datasets are the same as in TED-LIUM 2 (and TED-LIUM 1).
-- the 'speaker adaptation' one, especially designed for experiments on speaker adaptation.

---

SPH format info:

Channels       : 1
Sample Rate    : 16000
Precision      : 16-bit
Bit Rate       : 256k
Sample Encoding: 16-bit Signed Integer PCM

---

François Hernandez, Vincent Nguyen, Sahar Ghannay, Natalia Tomashenko, and Yannick Estève, "TED-LIUM 3: twice as much data and corpus repartition for experiments on speaker adaptation", submitted to the 20th International Conference on Speech and Computer (SPECOM 2018), September 2018, Leipzig, Germany
A preprint version is available on arxiv (and in the doc/ directory):
https://arxiv.org/abs/1805.04699
"""
import logging
import shutil
import tarfile
from concurrent.futures.thread import ThreadPoolExecutor
from functools import partial
from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from lhotse import (
    RecordingSet,
    SupervisionSegment,
    SupervisionSet,
    validate_recordings_and_supervisions,
)
from lhotse.qa import fix_manifests
from lhotse.recipes.utils import normalize_text_tedlium
from lhotse.utils import Pathlike, resumable_download, safe_extract

TEDLIUM_PARTS = ("train", "dev", "test")


def download_tedlium(
    target_dir: Pathlike = ".", force_download: Optional[bool] = False
) -> Path:
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    tar_path = target_dir / "TEDLIUM_release-3.tgz"
    corpus_dir = target_dir / "TEDLIUM_release-3"
    completed_detector = corpus_dir / ".completed"
    if completed_detector.is_file():
        logging.info(f"Skipping {tar_path.name} because {completed_detector} exists.")
        return corpus_dir
    resumable_download(
        "http://www.openslr.org/resources/51/TEDLIUM_release-3.tgz",
        filename=tar_path,
        force_download=force_download,
    )
    shutil.rmtree(corpus_dir, ignore_errors=True)
    with tarfile.open(tar_path) as tar:
        safe_extract(tar, path=target_dir)
    completed_detector.touch()
    return corpus_dir


def prepare_tedlium(
    tedlium_root: Pathlike,
    output_dir: Optional[Pathlike] = None,
    dataset_parts: Union[str, Sequence[str]] = TEDLIUM_PARTS,
    num_jobs: int = 1,
    normalize_text: str = "none",
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Prepare manifests for the TED-LIUM v3 corpus.

    The manifests are created in a dict with three splits: train, dev and test.
    Each split contains a RecordingSet and SupervisionSet in a dict under keys 'recordings' and 'supervisions'.

    :param tedlium_root: Path to the unpacked TED-LIUM data.
    :param output_dir: Path where the manifests should be written.
    :param dataset_parts: Which parts of the dataset to prepare.
        By default, all parts are prepared.
    :param num_jobs: Number of parallel jobs to use.
    :return: A dict with standard corpus splits containing the manifests.
    """
    tedlium_root = Path(tedlium_root)
    output_dir = Path(output_dir) if output_dir is not None else None
    corpus = {}

    dataset_parts = [dataset_parts] if isinstance(dataset_parts, str) else dataset_parts

    with ThreadPoolExecutor(num_jobs) as ex:
        for split in dataset_parts:
            logging.info(f"Processing {split} split...")
            root = tedlium_root / "legacy" / split
            recordings = RecordingSet.from_dir(
                root / "sph", pattern="*.sph", num_jobs=num_jobs
            )
            stms = list((root / "stm").glob("*.stm"))
            assert len(stms) == len(recordings), (
                f"Mismatch: found {len(recordings)} "
                f"sphere files and {len(stms)} STM files. "
                f"You might be missing some parts of TEDLIUM..."
            )
            futures = []
            _parse_stm_worker = partial(_parse_stm_file, normalize_text=normalize_text)
            for stm in stms:
                futures.append(ex.submit(_parse_stm_worker, stm))

            segments = []
            for future in futures:
                segments.extend(future.result())

            supervisions = SupervisionSet.from_segments(segments)
            recordings, supervisions = fix_manifests(recordings, supervisions)

            corpus[split] = {"recordings": recordings, "supervisions": supervisions}
            validate_recordings_and_supervisions(**corpus[split])

            if output_dir is not None:
                recordings.to_file(output_dir / f"tedlium_recordings_{split}.jsonl.gz")
                supervisions.to_file(
                    output_dir / f"tedlium_supervisions_{split}.jsonl.gz"
                )

    return corpus


def _parse_stm_file(stm: str, normalize_text: str = "none") -> SupervisionSegment:
    """Helper function to parse a single STM file."""
    segments = []
    with stm.open() as f:
        for idx, l in enumerate(f):
            rec_id, _, _, start, end, _, *words = l.split()
            start, end = float(start), float(end)
            text = " ".join(words).replace("{NOISE}", "[NOISE]")
            if text == "ignore_time_segment_in_scoring":
                continue
            segments.append(
                SupervisionSegment(
                    id=f"{rec_id}-{idx}",
                    recording_id=rec_id,
                    start=start,
                    duration=round(end - start, ndigits=8),
                    channel=0,
                    text=normalize_text_tedlium(text, normalize_text),
                    language="English",
                    speaker=rec_id,
                )
            )
    return segments



import bisect
import math
import random
from typing import Any, Dict, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import torch

from lhotse import CutSet, FeatureExtractor
from lhotse.augmentation import dereverb_wpe_torch
from lhotse.utils import Pathlike

__all__ = ["GlobalMVN", "SpecAugment", "RandomizedSmoothing", "DereverbWPE"]


class GlobalMVN(torch.nn.Module):
    """Apply global mean and variance normalization"""

    def __init__(self, feature_dim: int):
        super().__init__()
        self.feature_dim = feature_dim
        self.register_buffer("norm_means", torch.zeros(feature_dim))
        self.register_buffer("norm_stds", torch.ones(feature_dim))

    @classmethod
    def from_cuts(
        cls,
        cuts: CutSet,
        max_cuts: Optional[int] = None,
        extractor: Optional[FeatureExtractor] = None,
    ) -> "GlobalMVN":
        stats = cuts.compute_global_feature_stats(
            max_cuts=max_cuts, extractor=extractor
        )
        stats = {name: torch.as_tensor(value) for name, value in stats.items()}
        (feature_dim,) = stats["norm_means"].shape
        global_mvn = cls(feature_dim)
        global_mvn.load_state_dict(stats)
        return global_mvn

    @classmethod
    def from_file(cls, stats_file: Pathlike) -> "GlobalMVN":
        stats = torch.load(stats_file)
        (feature_dim,) = stats["norm_means"].shape
        global_mvn = cls(feature_dim)
        global_mvn.load_state_dict(stats)
        return global_mvn

    def to_file(self, stats_file: Pathlike):
        torch.save(self.state_dict(), stats_file)

    def forward(
        self,
        features: torch.Tensor,
        supervision_segments: Optional[torch.IntTensor] = None,
    ) -> torch.Tensor:
        return (features - self.norm_means) / self.norm_stds

    def inverse(self, features: torch.Tensor) -> torch.Tensor:
        return features * self.norm_stds + self.norm_means


class RandomizedSmoothing(torch.nn.Module):
    """
    Randomized smoothing - gaussian noise added to an input waveform, or a batch of waveforms.
    The summed audio is clipped to ``[-1.0, 1.0]`` before returning.
    """

    def __init__(
        self,
        sigma: Union[float, Sequence[Tuple[int, float]]] = 0.1,
        sample_sigma: bool = True,
        p: float = 0.3,
    ):
        """
        RandomizedSmoothing's constructor.

        :param sigma: standard deviation of the gaussian noise. Either a constant float, or a schedule,
            i.e. a list of tuples that specify which value to use from which step.
            For example, ``[(0, 0.01), (1000, 0.1)]`` means that from steps 0-999, the sigma value
            will be 0.01, and from step 1000 onwards, it will be 0.1.
        :param sample_sigma: when ``False``, then sigma is used as the standard deviation in each forward step.
            When ``True``, the standard deviation is sampled from a uniform distribution of
            ``[-sigma, sigma]`` for each forward step.
        :param p: the probability of applying this transform.
        """
        super().__init__()
        self.sigma = sigma
        self.sample_sigma = sample_sigma
        self.p = p
        self.step = 0

    def forward(self, audio: torch.Tensor, *args, **kwargs) -> torch.Tensor:

        # Determine the stddev value
        if isinstance(self.sigma, float):
            # Use a constant stddev value
            sigma = self.sigma
        else:
            # Determine the right stddev value from a given schedule.
            sigma = schedule_value_for_step(self.sigma, self.step)
            self.step += 1

        if self.sample_sigma:
            # In this mode stddev is stochastic itself
            # and is sampled from uniform distribution bounded by [-sigma, sigma] .
            mask_shape = (audio.shape[0],) + tuple(1 for _ in audio.shape[1:])
            # Sigma is of shape (batch_size, 1) - different for each noise example.
            sigma = sigma * (2 * torch.rand(mask_shape) - 1)

        # Create the random noise examples with identical sigma's.
        noise = sigma * torch.randn_like(audio)

        # Apply the transform with a probability p -> mask noise examples with probability 1 - p.
        noise_mask = random_mask_along_batch_axis(noise, p=1.0 - self.p)
        noise = noise * noise_mask

        return torch.clip(audio + noise, min=-1.0, max=1.0)


class SpecAugment(torch.nn.Module):
    """
    SpecAugment performs three augmentations:
    - time warping of the feature matrix
    - masking of ranges of features (frequency bands)
    - masking of ranges of frames (time)

    The current implementation works with batches, but processes each example separately
    in a loop rather than simultaneously to achieve different augmentation parameters for
    each example.
    """

    def __init__(
        self,
        time_warp_factor: Optional[int] = 80,
        num_feature_masks: int = 2,
        features_mask_size: int = 27,
        num_frame_masks: int = 10,
        frames_mask_size: int = 100,
        max_frames_mask_fraction: float = 0.15,
        p=0.9,
    ):
        """
        SpecAugment's constructor.

        :param time_warp_factor: parameter for the time warping; larger values mean more warping.
            Set to ``None``, or less than ``1``, to disable.
        :param num_feature_masks: how many feature masks should be applied. Set to ``0`` to disable.
        :param features_mask_size: the width of the feature mask (expressed in the number of masked feature bins).
            This is the ``F`` parameter from the SpecAugment paper.
        :param num_frame_masks: the number of masking regions for utterances. Set to ``0`` to disable.
        :param frames_mask_size: the width of the frame (temporal) masks (expressed in the number of masked frames).
            This is the ``T`` parameter from the SpecAugment paper.
        :param max_frames_mask_fraction: limits the size of the frame (temporal) mask to this value times the length
            of the utterance (or supervision segment).
            This is the parameter denoted by ``p`` in the SpecAugment paper.
        :param p: the probability of applying this transform.
            It is different from ``p`` in the SpecAugment paper!
        """
        super().__init__()
        assert 0 <= p <= 1
        assert num_feature_masks >= 0
        assert num_frame_masks >= 0
        assert features_mask_size > 0
        assert frames_mask_size > 0
        self.time_warp_factor = time_warp_factor
        self.num_feature_masks = num_feature_masks
        self.features_mask_size = features_mask_size
        self.num_frame_masks = num_frame_masks
        self.frames_mask_size = frames_mask_size
        self.max_frames_mask_fraction = max_frames_mask_fraction
        self.p = p

    def forward(
        self,
        features: torch.Tensor,
        supervision_segments: Optional[torch.IntTensor] = None,
        *args,
        **kwargs,
    ) -> torch.Tensor:
        """
        Computes SpecAugment for a batch of feature matrices.

        Since the batch will usually already be padded, the user can optionally
        provide a ``supervision_segments`` tensor that will be used to apply SpecAugment
        only to selected areas of the input. The format of this input is described below.

        :param features: a batch of feature matrices with shape ``(B, T, F)``.
        :param supervision_segments: an int tensor of shape ``(S, 3)``. ``S`` is the number of
            supervision segments that exist in ``features`` -- there may be either
            less or more than the batch size.
            The second dimension encoder three kinds of information:
            the sequence index of the corresponding feature matrix in `features`,
            the start frame index, and the number of frames for each segment.
        :return: an augmented tensor of shape ``(B, T, F)``.
        """
        assert len(features.shape) == 3, (
            "SpecAugment only supports batches of " "single-channel feature matrices."
        )
        features = features.clone()
        if supervision_segments is None:
            # No supervisions - apply spec augment to full feature matrices.
            for sequence_idx in range(features.size(0)):
                features[sequence_idx] = self._forward_single(features[sequence_idx])
        else:
            # Supervisions provided - we will apply time warping only on the supervised areas.
            for sequence_idx, start_frame, num_frames in supervision_segments:
                end_frame = start_frame + num_frames
                features[sequence_idx, start_frame:end_frame] = self._forward_single(
                    features[sequence_idx, start_frame:end_frame], warp=True, mask=False
                )
            # ... and then time-mask the full feature matrices. Note that in this mode,
            # it might happen that masks are applied to different sequences/examples
            # than the time warping.
            for sequence_idx in range(features.size(0)):
                features[sequence_idx] = self._forward_single(
                    features[sequence_idx], warp=False, mask=True
                )
        return features

    def _forward_single(
        self, features: torch.Tensor, warp: bool = True, mask: bool = True
    ) -> torch.Tensor:
        """
        Apply SpecAugment to a single feature matrix of shape (T, F).
        """
        if random.random() > self.p:
            # Randomly choose whether this transform is applied
            return features
        if warp:
            if self.time_warp_factor is not None and self.time_warp_factor >= 1:
                features = time_warp(features, factor=self.time_warp_factor)
        if mask:
            mean = features.mean()
            # Frequency masking
            features = mask_along_axis_optimized(
                features,
                mask_size=self.features_mask_size,
                mask_times=self.num_feature_masks,
                mask_value=mean,
                axis=2,
            )
            # Time masking
            max_tot_mask_frames = self.max_frames_mask_fraction * features.size(0)
            num_frame_masks = min(
                self.num_frame_masks,
                math.ceil(max_tot_mask_frames / self.frames_mask_size),
            )
            max_mask_frames = min(
                self.frames_mask_size, max_tot_mask_frames // num_frame_masks
            )
            features = mask_along_axis_optimized(
                features,
                mask_size=max_mask_frames,
                mask_times=num_frame_masks,
                mask_value=mean,
                axis=1,
            )

        return features

    def state_dict(self, **kwargs) -> Dict[str, Any]:
        return dict(
            time_warp_factor=self.time_warp_factor,
            num_feature_masks=self.num_feature_masks,
            features_mask_size=self.features_mask_size,
            num_frame_masks=self.num_frame_masks,
            frames_mask_size=self.frames_mask_size,
            max_frames_mask_fraction=self.max_frames_mask_fraction,
            p=self.p,
        )

    def load_state_dict(self, state_dict: Dict[str, Any]):
        self.time_warp_factor = state_dict.get(
            "time_warp_factor", self.time_warp_factor
        )
        self.num_feature_masks = state_dict.get(
            "num_feature_masks", self.num_feature_masks
        )
        self.features_mask_size = state_dict.get(
            "features_mask_size", self.features_mask_size
        )
        self.num_frame_masks = state_dict.get("num_frame_masks", self.num_frame_masks)
        self.frames_mask_size = state_dict.get(
            "frames_mask_size", self.frames_mask_size
        )
        self.max_frames_mask_fraction = state_dict.get(
            "max_frames_mask_fraction", self.max_frames_mask_fraction
        )
        self.p = state_dict.get("p", self.p)


def mask_along_axis_optimized(
    features: torch.Tensor,
    mask_size: int,
    mask_times: int,
    mask_value: float,
    axis: int,
) -> torch.Tensor:
    """
    Apply Frequency and Time masking along axis.
    Frequency and Time masking as described in the SpecAugment paper.

    :param features: input tensor of shape ``(T, F)``
    :mask_size: the width size for masking.
    :mask_times: the number of masking regions.
    :mask_value: Value to assign to the masked regions.
    :axis: Axis to apply masking on (1 -> time, 2 -> frequency)
    """
    if axis not in [1, 2]:
        raise ValueError("Only Frequency and Time masking are supported!")

    features = features.unsqueeze(0)
    features = features.reshape([-1] + list(features.size()[-2:]))

    values = torch.randint(int(0), int(mask_size), (1, mask_times))
    min_values = torch.rand(1, mask_times) * (features.size(axis) - values)
    mask_starts = (min_values.long()).squeeze()
    mask_ends = (min_values.long() + values.long()).squeeze()

    if axis == 1:
        if mask_times == 1:
            features[:, mask_starts:mask_ends] = mask_value
            return features.squeeze(0)
        for (mask_start, mask_end) in zip(mask_starts, mask_ends):
            features[:, mask_start:mask_end] = mask_value
    else:
        if mask_times == 1:
            features[:, :, mask_starts:mask_ends] = mask_value
            return features.squeeze(0)
        for (mask_start, mask_end) in zip(mask_starts, mask_ends):
            features[:, :, mask_start:mask_end] = mask_value

    features = features.squeeze(0)
    return features


def time_warp(features: torch.Tensor, factor: int) -> torch.Tensor:
    """
    Time warping as described in the SpecAugment paper.
    Implementation based on Espresso:
    https://github.com/freewym/espresso/blob/master/espresso/tools/specaug_interpolate.py#L51

    :param features: input tensor of shape ``(T, F)``
    :param factor: time warping parameter.
    :return: a warped tensor of shape ``(T, F)``
    """
    t = features.size(0)
    if t - factor <= factor + 1:
        return features
    center = np.random.randint(factor + 1, t - factor)
    warped = np.random.randint(center - factor, center + factor + 1)
    if warped == center:
        return features
    features = features.unsqueeze(0).unsqueeze(0)
    left = torch.nn.functional.interpolate(
        features[:, :, :center, :],
        size=(warped, features.size(3)),
        mode="bicubic",
        align_corners=False,
    )
    right = torch.nn.functional.interpolate(
        features[:, :, center:, :],
        size=(t - warped, features.size(3)),
        mode="bicubic",
        align_corners=False,
    )
    return torch.cat((left, right), dim=2).squeeze(0).squeeze(0)


T = TypeVar("T")


def schedule_value_for_step(schedule: Sequence[Tuple[int, T]], step: int) -> T:
    milestones, values = zip(*schedule)
    assert milestones[0] <= step, (
        f"Cannot determine the scheduled value for step {step} with schedule: {schedule}. "
        f"Did you forget to add the first part of the schedule "
        f"for steps below {milestones[0]}?"
    )
    idx = bisect.bisect_right(milestones, step) - 1
    return values[idx]


def random_mask_along_batch_axis(tensor: torch.Tensor, p: float = 0.5) -> torch.Tensor:
    """
    For a given tensor with shape (N, d1, d2, d3, ...), returns a mask with shape (N, 1, 1, 1, ...),
    that randomly masks the samples in a batch.

    E.g. for a 2D input matrix it looks like:

        >>> [[0., 0., 0., ...],
        ...  [1., 1., 1., ...],
        ...  [0., 0., 0., ...]]

    :param tensor: the input tensor.
    :param p: the probability of masking an element.
    """
    mask_shape = (tensor.shape[0],) + tuple(1 for _ in tensor.shape[1:])
    mask = (torch.rand(mask_shape) > p).to(torch.float32)
    return mask


class DereverbWPE(torch.nn.Module):
    """
    Dereverberation with Weighted Prediction Error (WPE).
    The implementation and default values are borrowed from `nara_wpe` package:
    https://github.com/fgnt/nara_wpe

    The method and library are described in the following paper:
    https://groups.uni-paderborn.de/nt/pubs/2018/ITG_2018_Drude_Paper.pdf
    """

    def __init__(self, n_fft: int = 512, hop_length: int = 128):
        super().__init__()
        self.n_fft = n_fft
        self.hop_length = hop_length

    def forward(self, audio: torch.Tensor, *args, **kwargs) -> torch.Tensor:
        """
        Expects audio to be 2D or 3D tensor.
        2D means a batch of single-channel audio, shape (B, T).
        3D means a batch of multi-channel audio, shape (B, D, T).
        B => batch size; D => number of channels; T => number of audio samples.
        """

        # Assume batch of single-channel data: apply dereverb to each example independently.
        if audio.ndim == 2:
            return torch.cat(
                [
                    dereverb_wpe_torch(
                        a.unsqueeze(0), n_fft=self.n_fft, hop_length=self.hop_length
                    )
                    for a in audio
                ],
                dim=0,
            )

        # Assume batch of multi-channel data: each example has D channels.
        assert audio.ndim == 3
        return torch.stack(
            [
                dereverb_wpe_torch(a, n_fft=self.n_fft, hop_length=self.hop_length)
                for a in audio
            ],
            dim=0,
        )



"""
This is a data preparation recipe for the National Corpus of Speech in Singaporean English.

The entire corpus is organised into a few parts.

Part 1 features about 1000 hours of prompted recordings of phonetically-balanced scripts from about 1000 local English speakers.

Part 2 presents about 1000 hours of prompted recordings of sentences randomly generated from words based on people, food, location, brands, etc, from about 1000 local English speakers as well. Transcriptions of the recordings have been done orthographically and are available for download.

Part 3 consists of about 1000 hours of conversational data recorded from about 1000 local English speakers, split into pairs. The data includes conversations covering daily life and of speakers playing games provided.

Parts 1 and 2 were recorded in quiet rooms using 3 microphones: a headset/ standing microphone (channel 0), a boundary microphone (channel 1), and a mobile phone (channel 3). Recordings that are available for download here have been down-sampled to 16kHz. Details of the microphone models used for each speaker as well as some corresponding non-personal and anonymized information can be found in the accompanying spreadsheets.

Part 3's recordings were split into 2 environments. In the Same Room environment where speakers were in same room, the recordings were done using 2 microphones: a close-talk mic and a boundary mic. In the Separate Room environment, speakers were separated into individual rooms. The recordings were done using 2 microphones in each room: a standing mic and a telephone.

Under Part 4, speakers were encouraged as best as possible to switch from Singapore English to their Mother Tongue languages. These recordings were done under two environments. In the Same Room recordings, speakers sit at least two metres apart and record using their mobile phones. In the Different Room environment, speakers would speak through each other via Zoom on their laptops, and recording using their mobile phones.

Under Part 5, speakers were made to speak following the 4 styles: Debate, Finance topics, Positive Emotion and Negative Emotions. All recordings were done in a Separate room session, via Zoom, where the audio is recorded using the mobile phone.

Under Part 6, speakers were made to speak following the 3 styles within either of the 3 designs: Design 1 (holiday/hotel/restaurant), Design 2 (bank, telephone, insurance), Design 3 (HDB, MOE, MSF). All recordings were done in a Separate room session, via Zoom, where the audio is recorded using the mobile phone.

We currently only support the part 3 recordings, in "same room close mic" and "separate rooms phone mic" environments.
"""
import itertools
import logging
import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from lhotse import (
    Recording,
    RecordingSet,
    SupervisionSegment,
    SupervisionSet,
    validate_recordings_and_supervisions,
)
from lhotse.parallel import parallel_map
from lhotse.qa import fix_manifests
from lhotse.utils import Pathlike

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

NSC_PARTS = [
    "PART1_CHANNEL0",
    "PART1_CHANNEL1",
    "PART1_CHANNEL2",
    "PART2_CHANNEL0",
    "PART2_CHANNEL1",
    "PART2_CHANNEL2",
    "PART3_SameBoundaryMic",
    "PART3_SameCloseMic",
    "PART3_SeparateIVR",
    "PART3_SeparateStandingMic",
    "PART4_CodeswitchingDiffRoom",
    "PART4_CodeswitchingSameRoom",
    "PART5_Debate",
    "PART5_FinanceEmotion",
    "PART6_CallCentreDesign1",
    "PART6_CallCentreDesign2",
    "PART6_CallCentreDesign3",
]


@dataclass
class ScriptAudioDir:
    script_dir: Union[str, Path]
    audio_dir: Union[str, Path]

    def relative_to(self, parent: Union[str, Path]) -> "ScriptAudioDir":
        parent = Path(parent)
        return ScriptAudioDir(
            script_dir=parent / self.script_dir,
            audio_dir=parent / self.audio_dir,
        )


@dataclass
class HandlerMapping:
    handler: Callable[[str, ScriptAudioDir, int], dict]
    script_audio: ScriptAudioDir


# function forward declaration is not supported, wrap static data into function
def get_part_handler_map(corpus_dir: Path) -> Dict[str, HandlerMapping]:
    part_1_3_parent_dir = corpus_dir / "IMDA - National Speech Corpus"
    part_4_6_parent_dir = (
        corpus_dir
        / "IMDA - National Speech Corpus - Additional"
        / "IMDA - National Speech Corpus (Additional)"
    )
    # fmt: off
    part_handler_mapping: Dict[str, HandlerMapping] = {
        "PART1_CHANNEL0": HandlerMapping(handler=prepare_part1, script_audio=ScriptAudioDir(script_dir="PART1/DATA/CHANNEL0/SCRIPT", audio_dir="PART1/DATA/CHANNEL0/WAVE").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART1_CHANNEL1": HandlerMapping(handler=prepare_part1, script_audio=ScriptAudioDir(script_dir="PART1/DATA/CHANNEL1/SCRIPT", audio_dir="PART1/DATA/CHANNEL1/WAVE").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART1_CHANNEL2": HandlerMapping(handler=prepare_part1, script_audio=ScriptAudioDir(script_dir="PART1/DATA/CHANNEL2/SCRIPT", audio_dir="PART1/DATA/CHANNEL2/WAVE").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART2_CHANNEL0": HandlerMapping(handler=prepare_part2, script_audio=ScriptAudioDir(script_dir="PART2/DATA/CHANNEL0/SCRIPT", audio_dir="PART2/DATA/CHANNEL0/WAVE").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART2_CHANNEL1": HandlerMapping(handler=prepare_part2, script_audio=ScriptAudioDir(script_dir="PART2/DATA/CHANNEL1/SCRIPT", audio_dir="PART2/DATA/CHANNEL1/WAVE").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART2_CHANNEL2": HandlerMapping(handler=prepare_part2, script_audio=ScriptAudioDir(script_dir="PART2/DATA/CHANNEL2/SCRIPT", audio_dir="PART2/DATA/CHANNEL2/WAVE").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART3_SameBoundaryMic": HandlerMapping(handler=prepare_part3, script_audio=ScriptAudioDir(script_dir="PART3/Scripts Same", audio_dir="PART3/Audio Same BoundaryMic").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART3_SameCloseMic": HandlerMapping(handler=prepare_part3,script_audio= ScriptAudioDir(script_dir="PART3/Scripts Same", audio_dir="PART3/Audio Same CloseMic").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART3_SeparateIVR": HandlerMapping(handler=prepare_part3, script_audio=ScriptAudioDir(script_dir="PART3/Scripts Separate", audio_dir="PART3/Audio Separate IVR").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART3_SeparateStandingMic": HandlerMapping(handler=prepare_part3,script_audio=ScriptAudioDir(script_dir="PART3/Scripts Separate", audio_dir="PART3/Audio Separate StandingMic").relative_to(part_1_3_parent_dir)), # noqa: E501
        "PART4_CodeswitchingDiffRoom": HandlerMapping(handler=prepare_part4, script_audio=ScriptAudioDir(script_dir="PART4/Codeswitching/Diff Room Scripts", audio_dir="PART4/Codeswitching/Diff Room Audio").relative_to(part_4_6_parent_dir)), # noqa: E501
        "PART4_CodeswitchingSameRoom": HandlerMapping(handler=prepare_part4, script_audio=ScriptAudioDir(script_dir="PART4/Codeswitching/Same Room Scripts", audio_dir="PART4/Codeswitching/Same Room Audio").relative_to(part_4_6_parent_dir)), # noqa: E501
        "PART5_Debate": HandlerMapping(handler=prepare_part5, script_audio=ScriptAudioDir(script_dir="PART5/Debate Scripts", audio_dir="PART5/Debate Audio").relative_to(part_4_6_parent_dir)), # noqa: E501
        "PART5_FinanceEmotion": HandlerMapping(handler=prepare_part5, script_audio=ScriptAudioDir(script_dir="PART5/Finance + Emotion Scripts", audio_dir="PART5/Finance + Emotions Audio").relative_to(part_4_6_parent_dir)), # noqa: E501
        "PART6_CallCentreDesign1": HandlerMapping(handler=prepare_part6, script_audio=ScriptAudioDir(script_dir="PART6/Call Centre Design 1/Scripts", audio_dir="PART6/Call Centre Design 1/Audio").relative_to(part_4_6_parent_dir)), # noqa: E501
        "PART6_CallCentreDesign2": HandlerMapping(handler=prepare_part6, script_audio=ScriptAudioDir(script_dir="PART6/Call Centre Design 2/Scripts", audio_dir="PART6/Call Centre Design 2/Audio").relative_to(part_4_6_parent_dir)), # noqa: E501
        "PART6_CallCentreDesign3": HandlerMapping(handler=prepare_part6, script_audio=ScriptAudioDir(script_dir="PART6/Call Centre Design 3/Scripts", audio_dir="PART6/Call Centre Design 3/Audio").relative_to(part_4_6_parent_dir)), # noqa: E501
    }
    return part_handler_mapping
    # fmt: on


def check_dependencies():
    try:
        import textgrids  # noqa
    except:
        raise ImportError(
            "NSC data preparation requires the forked 'textgrids' package to be installed. "
            "Please install it with 'pip install git+https://github.com/pzelasko/Praat-textgrids' "
            "and try again."
        )


def prepare_nsc(
    corpus_dir: Pathlike,
    dataset_part: str = "PART3_SameCloseMic",
    output_dir: Optional[Pathlike] = None,
    num_jobs=1,
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    :param corpus_dir: Pathlike, the path to the raw corpus distribution.
    :param dataset_part: str, name of the dataset part to be prepared.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """
    check_dependencies()
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    part_handler_map = get_part_handler_map(corpus_dir)
    if dataset_part in part_handler_map:
        handler_map = part_handler_map[dataset_part]
        manifests = handler_map.handler(
            dataset_part,
            handler_map.script_audio,
            num_jobs,
        )
    else:
        raise ValueError(f"Unknown dataset part: {dataset_part}")

    # Fix the manifests to make sure they are valid
    recordings, supervisions = fix_manifests(**manifests)
    validate_recordings_and_supervisions(recordings, supervisions)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        supervisions.to_file(output_dir / f"nsc_supervisions_{dataset_part}.jsonl.gz")
        recordings.to_file(output_dir / f"nsc_recordings_{dataset_part}.jsonl.gz")

    return manifests


def prepare_part1(
    part_name: str,
    script_audio_dir: ScriptAudioDir,
    num_jobs: int = 1,
    *args,
    **kwargs,
):
    recordings = []
    supervisions = []
    audio_dir = Path(script_audio_dir.audio_dir)
    script_dir = Path(script_audio_dir.script_dir)
    channel = int(part_name[-1:])  # E.g. PART1_CHANNEL0
    assert channel in {0, 1, 2}
    extract_to_dir = audio_dir / "extracted"
    if not extract_to_dir.exists():
        extract_to_dir.mkdir()
    speaker_zip_files = [f for f in audio_dir.glob("SPEAKER*.zip")]
    for speaker_manifests in tqdm(
        parallel_map(
            _parse_part1_speaker,
            speaker_zip_files,
            itertools.repeat(script_dir),
            itertools.repeat(channel),
            itertools.repeat(extract_to_dir),
            num_jobs=num_jobs,
        ),
        total=len(speaker_zip_files),
        desc=f"Creating manifests for {part_name}",
    ):
        recordings.extend(speaker_manifests["recordings"])
        supervisions.extend(speaker_manifests["supervisions"])
    return {
        "recordings": RecordingSet.from_recordings(recordings),
        "supervisions": SupervisionSet.from_segments(supervisions),
    }


def prepare_part2(
    part_name: str,
    script_audio_dir: ScriptAudioDir,
    num_jobs: int = 1,
    *args,
    **kwargs,
):
    return prepare_part1(
        part_name=part_name,
        script_audio_dir=script_audio_dir,
        num_jobs=num_jobs,
        *args,
        **kwargs,
    )


def prepare_part3(
    part_name: str,
    script_audio_dir: ScriptAudioDir,
    num_jobs: int = 1,
    *args,
    **kwargs,
):
    from textgrids import TextGrid

    assert (
        part_name != "PART3_SameBoundaryMic"
    ), "The recipe too different, currently not supported"

    def textgrid_field_id_resolver(audio_file: Union[str, Path]) -> Tuple[Dict, str]:
        audio_file = Path(audio_file)
        script_dir = Path(script_audio_dir.script_dir)
        if part_name == "PART3_SeparateIVR":
            script_file_stem = audio_file.parent.name + "_" + audio_file.stem
            script_file = script_dir / f"{script_file_stem}.TextGrid"
            textgrid_key = script_file_stem
        else:
            script_file = script_dir / f"{audio_file.stem}.TextGrid"
            textgrid_key = audio_file.stem
        coding = _detect_textgrid_encoding(script_file)
        tg = TextGrid(script_file, coding=coding)
        return tg, textgrid_key

    return prepare_textgrid_based_part(
        part_name=part_name,
        script_audio_dir=script_audio_dir,
        textgrid_loader=textgrid_field_id_resolver,
        num_jobs=num_jobs,
        *args,
        **kwargs,
    )


def prepare_part4(
    part_name: str,
    script_audio_dir: ScriptAudioDir,
    num_jobs: int = 1,
    *args,
    **kwargs,
):
    from textgrids import TextGrid

    def textgrid_field_id_resolver(
        audio_file: Union[str, Path]
    ) -> Tuple[TextGrid, str]:
        audio_file = Path(audio_file)
        script_dir = Path(script_audio_dir.script_dir)
        script_file = script_dir / f"{audio_file.stem}.TextGrid"
        coding = _detect_textgrid_encoding(script_file)
        tg = TextGrid(script_file, coding=coding)
        # NSC textgrid change unpredictably, getting it from dict key
        textgrid_key = next(iter(tg.keys()))
        return tg, textgrid_key

    return prepare_textgrid_based_part(
        part_name=part_name,
        script_audio_dir=script_audio_dir,
        textgrid_loader=textgrid_field_id_resolver,
        num_jobs=num_jobs,
        *args,
        **kwargs,
    )


def prepare_part5(
    part_name: str,
    script_audio_dir: ScriptAudioDir,
    num_jobs: int = 1,
    *args,
    **kwargs,
):
    from textgrids import TextGrid

    def textgrid_field_id_resolver(
        audio_file: Union[str, Path]
    ) -> Tuple[TextGrid, str]:
        audio_file = Path(audio_file)
        script_dir = Path(script_audio_dir.script_dir)
        script_file = script_dir / f"{audio_file.stem}.TextGrid"
        coding = _detect_textgrid_encoding(script_file)
        tg = TextGrid(script_file, coding=coding)
        # NSC textgrid change unpredictably, getting it from dict key
        textgrid_key = next(iter(tg.keys()))
        return tg, textgrid_key

    return prepare_textgrid_based_part(
        part_name=part_name,
        script_audio_dir=script_audio_dir,
        textgrid_loader=textgrid_field_id_resolver,
        num_jobs=num_jobs,
        *args,
        **kwargs,
    )


def prepare_part6(
    part_name: str,
    script_audio_dir: ScriptAudioDir,
    num_jobs: int = 1,
    *args,
    **kwargs,
):
    return prepare_part5(
        part_name=part_name,
        script_audio_dir=script_audio_dir,
        num_jobs=num_jobs,
        *args,
        **kwargs,
    )


def prepare_textgrid_based_part(
    part_name: str,
    script_audio_dir: ScriptAudioDir,
    textgrid_loader: Callable[[Union[str, Path]], Tuple[Dict, str]],
    num_jobs: int = 1,
    *args,
    **kwargs,
):
    """Prepare part that use textgrid to storing script like: PART3, PART4, PART5, PART6

    Args:
        part_path (Path): path to part
        script_audio_dir (Path): root dir of audios
        script_file_resolver (Callable): a function resolve an audio path in to textgrid script file and it's key
        num_jobs (int): number of workers to process data
    """
    check_dependencies()

    recordings = []
    supervisions = []
    audio_dir = Path(script_audio_dir.audio_dir)
    audio_files = [
        f
        for f in itertools.chain(
            audio_dir.rglob("**/*.wav"), audio_dir.rglob("**/*.WAV")
        )
    ]
    processed_recordings = set()
    for audio_path in tqdm(
        audio_files,
        total=len(audio_files),
        desc=f"Creating manifests for {part_name}",
    ):
        try:
            recording_id = f"{part_name}_{audio_path.stem}"
            assert (
                recording_id not in processed_recordings
            ), f'Duplicated recording id "{recording_id}", audio path: "{audio_path}"'
            processed_recordings.add(recording_id)
            recording = Recording.from_file(audio_path, recording_id=recording_id)
            tg, textgrid_key = textgrid_loader(audio_path)
            segments = [
                s
                for s in (
                    SupervisionSegment(
                        id=f"{recording.id}-{idx}",
                        recording_id=recording.id,
                        start=segment.xmin,
                        # We're trimming the last segment's duration as it exceeds the actual duration of the recording.
                        # This is safe because if we end up with a zero/negative duration, the validation will catch it.
                        duration=min(
                            round(segment.xmax - segment.xmin, ndigits=8),
                            recording.duration - segment.xmin,
                        ),
                        text=segment.text,
                        language="Singaporean English",
                        speaker=recording_id,
                    )
                    for idx, segment in enumerate(tg[textgrid_key])
                    if segment.text not in ("<S>", "<Z>")  # skip silences
                )
                if s.duration > 0  # NSC has some bad segments
            ]

            supervisions.extend(segments)
            recordings.append(recording)
        except Exception:
            with logging_redirect_tqdm():
                logger.warning(f'Error when processing "{audio_path}" - skipping...')
    return {
        "recordings": RecordingSet.from_recordings(recordings),
        "supervisions": SupervisionSet.from_segments(supervisions),
    }


def _parse_part1_speaker(
    speaker_zip_file: Pathlike,
    script_dir: Pathlike,
    channel: int,
    extract_to_dir: Optional[Path] = None,
):
    script_session_dir_mapping = _preprocess_part1_speaker(
        speaker_zip_file=speaker_zip_file,
        script_dir=script_dir,
        channel=channel,
        extract_to_dir=extract_to_dir,
    )
    sessions_dir = []
    scripts_file = []
    for script_file, session_dir in script_session_dir_mapping.items():
        scripts_file.append(script_file)
        sessions_dir.append(session_dir)
    recordings: List[Recording] = []
    supervisions: List[SupervisionSegment] = []
    for r, s in [
        _parse_part1_script(sc_f, ss_d)
        for sc_f, ss_d in zip(scripts_file, sessions_dir)
    ]:
        recordings.extend(r)
        supervisions.extend(s)

    return {
        "recordings": RecordingSet.from_recordings(recordings),
        "supervisions": SupervisionSet.from_segments(supervisions),
    }


def _preprocess_part1_speaker(
    speaker_zip_file: Pathlike,
    script_dir: Pathlike,
    channel: int,
    extract_to_dir: Optional[Path] = None,
) -> Dict[Path, Path]:
    """Extract PART1/PART2 speaker audio

    Args:
        speaker_zip_file (Pathlike): Path to speaker zipped audio file
        script_dir (Pathlike): Path to script dir of the channel
        channel (int): Channel of the PART, we can parse from $script_dir but it is not necessary
        extract_to_dir (Optional[Path]): Directory to extract zipped audio file, default to parent dir of $speaker_zip_file

    Returns:
        Dict[Path, Path]: Mapping of script file -> speaker's session dir
    """
    speaker_zip_file = Path(speaker_zip_file)
    script_dir = Path(script_dir)
    if extract_to_dir is None:
        extract_to_dir = speaker_zip_file.parent
    speaker_audio_dir = extract_to_dir / speaker_zip_file.stem
    if not speaker_audio_dir.exists():
        with zipfile.ZipFile(speaker_zip_file) as zf:
            zf.extractall(extract_to_dir)
    else:
        with logging_redirect_tqdm():
            logger.warning(
                f'Reusing "{speaker_audio_dir}" as extracted "{speaker_zip_file}"'
                " since it is existed already"
            )
    sessions_dir = [f for f in speaker_audio_dir.glob("SESSION*")]
    scripts_file: List[Path] = []
    for session_dir in sessions_dir:
        spk_id = speaker_audio_dir.stem.removeprefix("SPEAKER")
        session_number = session_dir.stem.removeprefix("SESSION")
        session_script_file = script_dir / f"{channel}{spk_id}{session_number}.TXT"
        scripts_file.append(session_script_file)
    return {sr_f: ss_d for sr_f, ss_d in zip(scripts_file, sessions_dir)}


def _parse_part1_script(
    script_file: Path, session_dir
) -> Tuple[List[Recording], List[SupervisionSegment]]:
    recordings = []
    supervisions = []

    with open(script_file, "r", encoding="utf-8-sig") as fr:
        previous_audio_id = ""
        previous_text = ""
        for line in fr:
            columns = line.rstrip("\n").split("\t")
            if previous_audio_id != "" and columns[0] != previous_audio_id:
                # empty id indicate normalized text of prior line
                if columns[0] == "":
                    previous_text = columns[1]
                recording, segment = _create_part1_single_record(
                    session_dir, previous_audio_id, previous_text
                )
                if recording:
                    recordings.append(recording)
                    supervisions.append(segment)
                previous_audio_id = previous_text = ""
            else:
                previous_audio_id = columns[0]
                previous_text = columns[1]
        if previous_audio_id:
            recording, segment = _create_part1_single_record(
                session_dir, previous_audio_id, previous_text
            )
            if recording:
                recordings.append(recording)
                supervisions.append(segment)
    return recordings, supervisions


def _create_part1_single_record(
    session_dir: Path, _audio_id: str, text: str
) -> Tuple[Optional[Recording], Optional[SupervisionSegment]]:
    audio_file = session_dir / f"{_audio_id}.WAV"
    try:
        recording = Recording.from_file(audio_file, recording_id=_audio_id)
        segment = SupervisionSegment(
            id=recording.id,
            recording_id=recording.id,
            start=0,
            duration=recording.duration,
            text=text,
        )
        return recording, segment
    except FileNotFoundError:
        with logging_redirect_tqdm():
            logger.warning(
                f'Recording audio of script "{_audio_id}" '
                f'can not be found in "{session_dir}"'
            )
    except Exception as e:
        with logging_redirect_tqdm():
            logger.error(f"Error occurred with {audio_file}", e)
    return None, None


def _detect_textgrid_encoding(textgrid_file: Pathlike) -> Optional["str"]:
    """_summary_

    Returns:
        str: encoding of the file or None if it is binary file or undetectable
    """
    import charset_normalizer

    textgrid_binary_mark = b"ooBinaryFile\x08TextGrid"
    with open(textgrid_file, "rb") as fr:
        checking_bytes = fr.read(10 * 2**10)  # 10kB
    if checking_bytes[: len(textgrid_binary_mark)] == textgrid_binary_mark:
        return None
    else:
        charset_match = charset_normalizer.from_bytes(checking_bytes).best()  # type: ignore
        if charset_match is None:
            return None  # the most widely used I know
        else:
            # prefer utf-8 over ascii
            return (
                charset_match.encoding if charset_match.encoding != "ascii" else "utf-8"
            )



"""
This script creates the BUT Reverb DB dataset, which is available at:
https://speech.fit.vutbr.cz/software/but-speech-fit-reverb-database

The following description is taken from the official website:

This is the first release of BUT Speech@FIT Reverb Database. The database is being built
with respect to collect a large number of various Room Impulse Responses,
Room environmental noises (or "silences"), Retransmitted speech (for ASR and SID testing),
and meta-data (positions of microphones, speakers etc.).

The goal is to provide speech community with a dataset for data enhancement and distant
microphone or microphone array experiments in ASR and SID.

The BUT Speech@FIT Reverb Dataset consists of 9 rooms:

Size [m x m x m]	Volume [m^3]	# RIRs	Ret.	Type	In RIR-Only set	In LibriSpeech-Only set
Q301	10.7x6.9x2.6	192	31 x 3	1	Office	Yes	Yes
L207	4.6x6.9x3.1	98	31 x 6	3	Office	Yes	Yes
L212	7.5x4.6x3.1	107	31 x 5	2	Office	Yes	Yes
L227	6.2x2.6x14.2	229	31 x 5	3	Stairs	Yes	Yes
R112	4.4x2.8x2.6*	~40	31 x 5	0	Hotel room	Yes	No
CR2	28.2x11.1x3.3	1033	31 x 4	0	Conf. room	Yes	No
E112	11.5x20.1x4.8*	~900	31 x 2	0	Lect. room	Yes	No
D105	17.2x22.8x6.9*	~2000	31 x 6	1	Lect. room	Yes	Yes
C236	7.0x4.1x3.6	102	31 x 10	0	Meeting room	Yes	No

We placed 31 microphones in both rooms. The source (a Hi-Fi loudspeaker) was placed on 5
positions in average. We measured RIRs (using exponential sine sweep method) for each
speaker position. Next we recorded environmental noise (silence). There was a radio at
background playing in one speaker position in the office.

All microphone positions are measured and stored in meta-files. We pre-calculated positions
of microphones and speakers in Cartesian and polar coordinates as absolute and relative (to the speaker).

The corpus can be cited as follows:
@ARTICLE{8717722,
  author={Szöke, Igor and Skácel, Miroslav and Mošner, Ladislav and Paliesek, Jakub and Černocký, Jan},
  journal={IEEE Journal of Selected Topics in Signal Processing},
  title={Building and evaluation of a real room impulse response dataset},
  year={2019},
  volume={13},
  number={4},
  pages={863-876},
  doi={10.1109/JSTSP.2019.2917582}}
"""
import logging
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from tqdm import tqdm

from lhotse import CutSet, Recording, RecordingSet
from lhotse.utils import Pathlike, resumable_download

BUT_REVERB_DB_URL = (
    "http://merlin.fit.vutbr.cz/ReverbDB/BUT_ReverbDB_rel_19_06_RIR-Only.tgz"
)


def download_but_reverb_db(
    target_dir: Pathlike = ".",
    url: Optional[str] = BUT_REVERB_DB_URL,
    force_download: Optional[bool] = False,
) -> Path:
    """
    Download and untar the BUT Reverb DB dataset.

    :param target_dir: Pathlike, the path of the dir to store the dataset.
    :param url: str, the url that downloads file called BUT_ReverbDB.tgz.
    :param force_download: bool, if True, download the archive even if it already exists.
    """
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    tgz_name = "BUT_ReverbDB.tgz"
    tgz_path = target_dir / tgz_name
    if tgz_path.exists() and not force_download:
        logging.info(f"Skipping {tgz_name} because file exists.")
    resumable_download(url, tgz_path, force_download=force_download)
    tgz_dir = target_dir / "BUT_ReverbDB"
    if not tgz_dir.exists():
        logging.info(f"Untarring {tgz_name}.")
        with tarfile.open(tgz_path) as tar:
            tar.extractall(path=target_dir)
    return tgz_dir


def prepare_but_reverb_db(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    parts: Sequence[str] = ("silence", "rir"),
) -> Dict[str, Dict[str, Union[RecordingSet, CutSet]]]:
    """
    Prepare the BUT Speech@FIT Reverb Database corpus.

    :param corpus_dir: Pathlike, the path of the dir to store the dataset.
    :param output_dir: Pathlike, the path of the dir to write the manifests.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if not parts:
        raise ValueError("No parts specified for manifest preparation.")
    if isinstance(parts, str):
        parts = [parts]

    recordings = defaultdict(list)
    for wav_file in tqdm(corpus_dir.rglob("*.wav")):
        part = wav_file.parent.name.lower()
        if part not in parts:
            continue
        # The file path is like:
        # VUT_FIT_C236/MicID01/SpkID01_20190503_S/01/RIR/IR_sweep_15s_45Hzto22kHz_FS16kHz.v00.wav
        # We convert it to the following for the recording ID:
        # VUT_FIT_C236-MicID01-SpkID01_20190503_S-01-v00
        room_id = wav_file.parent.parent.parent.parent.parent.stem
        mic_id = wav_file.parent.parent.parent.parent.stem
        spk_id = wav_file.parent.parent.parent.stem
        uid = wav_file.parent.parent.stem
        version = wav_file.stem.split(".")[-1]
        recording_id = f"{room_id}-{mic_id}-{spk_id}-{uid}-v{version}"
        recording = Recording.from_file(wav_file, recording_id=recording_id)
        recordings[part].append(recording)

    manifests = defaultdict(dict)
    for part in parts:
        manifests[part]["recordings"] = RecordingSet.from_recordings(recordings[part])

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        for part in parts:
            manifests[part]["recordings"].to_file(
                output_dir / f"but-reverb-db_{part}_recordings.jsonl.gz"
            )

    return manifests



import logging
from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, resumable_download

SPATIAL_LIBRISPEECH = ("train", "test")
BASE_URL = "https://docs-assets.developer.apple.com/ml-research/datasets/spatial-librispeech/v1"
META_DATA_URL = "https://docs-assets.developer.apple.com/ml-research/datasets/spatial-librispeech/v1/metadata.parquet"


def _download_and_save_audio(target_file: Pathlike, url: str):
    # Implementation from https://github.com/apple/ml-spatial-librispeech/pull/1/
    # Use the requests module to avoid the 403 forbidden error
    def _download_file(url: str) -> bytes:
        """This function downloads and returns the content of the given url
        Args:
            url (str): the url of the file to be downloaded
        Raises:
            e: The exception that is raised by the request module
        Returns:
            file_content (bytes): The file content downloaded from the url
        """

        try:
            import requests
        except ImportError:
            raise ImportError(
                "The Spatial LibriSpeech recipe requires requests dependency to download the dataset. You can install the dependency using: pip install requests"
            )

        try:
            file_content = requests.get(url, allow_redirects=True).content
            return file_content
        except requests.exceptions.RequestException as e:
            raise e

    # Implementation from https://github.com/apple/ml-spatial-librispeech/pull/1/
    def _save_audio_content(target_file: str, file_content: bytes):
        """This function saves the downloaded content passed via `file_content' in the `target_file'
        Args:
            target_file (str): the target path for the file content to be saved to
            file_content (bytes): the content to be saved

        Raises:
            e: the IOError raised by the writing operation
        """
        try:
            with open(target_file, "wb") as file:
                file.write(file_content)
        except IOError as e:
            raise e

    file_content = _download_file(url)
    _save_audio_content(target_file, file_content)


def download_spatial_librispeech(
    target_dir: Pathlike = ".",
    dataset_parts: Union[str, Sequence[str]] = SPATIAL_LIBRISPEECH,
    force_download: bool = False,
    base_url: str = BASE_URL,
    num_jobs: int = 1,
) -> Path:
    """
    Download the Spatial-LibriSpeech dataset.

    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param dataset_parts: "all" or a list of splits (e.g. ["train", "test"]) to download.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the url of the resource.
    :return: the path to downloaded and extracted directory with data.
    """

    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "The Spatial LibriSpeech recipe requires pandas, pyarrow and fastparquet dependency to parse parquet formatted metadata. You can install the dependencies using: pip install pandas pyarrow fastparquet"
        )

    def _download_spatial_librispeech_audio_files(
        target_dir: Pathlike,
        dataset_parts: Sequence[str],
        metadata: pd.DataFrame,
        base_url: str,
        force_download: bool = False,
        num_jobs: int = 1,
    ):
        target_dir = Path(target_dir)
        target_dir.mkdir(parents=True, exist_ok=True)

        audio_url = f"{base_url}/ambisonics"
        from concurrent.futures.thread import ThreadPoolExecutor

        for part in dataset_parts:
            part_dir = target_dir / part
            part_dir.mkdir(parents=True, exist_ok=True)

        with ThreadPoolExecutor(num_jobs) as ex:
            for sample_id, split in tqdm(
                zip(metadata["sample_id"], metadata["split"]),
                total=len(metadata["sample_id"]),
            ):
                if split not in dataset_parts:
                    continue
                recording_path = target_dir / split / f"{sample_id:06}.flac"
                recording_url = f"{audio_url}/{sample_id:06}.flac"
                if not recording_path.exists() or force_download:
                    ex.submit(_download_and_save_audio, recording_path, recording_url)

    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    if dataset_parts == "all":
        dataset_parts = SPATIAL_LIBRISPEECH
    else:
        dataset_parts = (
            [dataset_parts] if isinstance(dataset_parts, str) else dataset_parts
        )
    for part in dataset_parts:
        assert part in SPATIAL_LIBRISPEECH, f"Unknown dataset part: {part}"

    corpus_dir = target_dir / "Spatial-LibriSpeech"
    corpus_dir.mkdir(parents=True, exist_ok=True)

    completed_detector = corpus_dir / ".completed"
    if completed_detector.is_file():
        logging.info(f"Skipping download, found {completed_detector}.")
        return corpus_dir

    metadata_path = corpus_dir / "metadata.parquet"
    if not metadata_path.is_file() or force_download:
        resumable_download(META_DATA_URL, metadata_path, force_download=force_download)
    elif metadata_path.is_file():
        logging.info(f"Skipping download, found {metadata_path}.")

    metadata = pd.read_parquet(metadata_path)
    try:
        _download_spatial_librispeech_audio_files(
            target_dir=corpus_dir / "audio_files",
            dataset_parts=dataset_parts,
            metadata=metadata,
            base_url=base_url,
            force_download=force_download,
            num_jobs=num_jobs,
        )
    except Exception as e:
        logging.error(f"Failed to download audio files: {e}")
        raise e

    completed_detector.touch()
    return corpus_dir


def prepare_spatial_librispeech(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    dataset_parts: Union[str, Sequence[str]] = SPATIAL_LIBRISPEECH,
    normalize_text: str = "none",
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param dataset_parts: string or sequence of strings representing dataset part names, e.g. 'train', 'test'.
        By default we will infer which parts are available in ``corpus_dir``.
    :param normalize_text: str, "none" or "lower",
        for "lower" the transcripts are converted to lower-case.
    :param num_jobs: int, number of parallel threads used for 'parse_utterance' calls.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """

    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "The Spatial LibriSpeech recipe requires pandas, pyarrow and fastparquet dependency to parse parquet formatted metadata. You can install the dependencies using: pip install pandas pyarrow fastparquet"
        )

    corpus_dir = Path(corpus_dir)
    output_dir = Path(output_dir) if output_dir is not None else corpus_dir
    output_dir.mkdir(parents=True, exist_ok=True)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    if dataset_parts == "all":
        dataset_parts = SPATIAL_LIBRISPEECH
    else:
        dataset_parts = (
            [dataset_parts] if isinstance(dataset_parts, str) else dataset_parts
        )
    for part in dataset_parts:
        assert part in SPATIAL_LIBRISPEECH, f"Unknown dataset part: {part}"

    metadata_path = corpus_dir / "metadata.parquet"
    assert metadata_path.is_file(), f"{metadata_path} not found"
    metadata = pd.read_parquet(metadata_path)

    manifests = {}

    for part in dataset_parts:
        assert part in SPATIAL_LIBRISPEECH, f"Unknown dataset part: {part}"
        logging.info(f"Processing {part} split...")
        part_dir = corpus_dir / "audio_files" / part
        recording_set = RecordingSet.from_dir(
            part_dir,
            pattern="*.flac",
            num_jobs=num_jobs,
            recording_id=lambda x: x.stem,
        )

        supervision_segments = []
        part_metadata = metadata[metadata["split"] == part]
        for _, row in tqdm(
            part_metadata.iterrows(),
            total=len(part_metadata["sample_id"]),
            desc=f"Processing supervision segments for split: {part}",
        ):
            recording_id = f"{row['sample_id']:06}"
            start = 0
            duration = recording_set[recording_id].duration
            channel = recording_set[recording_id].channel_ids
            text = row["speech/librispeech_metadata/transcription"]
            speaker = row["speech/librispeech_metadata/reader_id"]
            gender = row["speech/librispeech_metadata/reader_sex"]
            segment = SupervisionSegment(
                id=recording_id,
                recording_id=recording_id,
                start=start,
                duration=duration,
                channel=channel,
                text=text,
                gender=gender,
                speaker=speaker,
            )
            supervision_segments.append(segment)
        supervision_set = SupervisionSet.from_segments(supervision_segments)

        # Normalize text to lowercase
        if normalize_text == "lower":
            to_lower = lambda text: text.lower()
            supervision_set = SupervisionSet.from_segments(
                [s.transform_text(to_lower) for s in supervision_set]
            )

        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            recording_set.to_file(
                output_dir / f"spatial-librispeech_recordings_{part}.jsonl.gz"
            )
            supervision_set.to_file(
                output_dir / f"spatial-librispeech_supervisions_{part}.jsonl.gz"
            )

        manifests[part] = {
            "recordings": recording_set,
            "supervisions": supervision_set,
        }

    return manifests



"""
LibriCSS is a multi-talker meeting corpus formed from mixing together LibriSpeech utterances
and replaying in a real meeting room. It consists of 10 1-hour sessions of audio, each
recorded on a 7-channel microphone. The sessions are recorded at a sampling rate of 16 kHz.
For more information, refer to the paper:
Z. Chen et al., "Continuous speech separation: dataset and analysis,"
ICASSP 2020 - 2020 IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP),
Barcelona, Spain, 2020
"""
import json
import logging
import subprocess
import zipfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Union

from tqdm import tqdm

from lhotse import (
    CutSet,
    RecordingSet,
    SupervisionSegment,
    SupervisionSet,
    fix_manifests,
    validate_recordings_and_supervisions,
)
from lhotse.audio import Recording
from lhotse.utils import Pathlike, fastcopy

# fmt: off
# The following mapping is courtesy Zhuo Chen (Microsoft). It is not available in the original
# LibriCSS dataset. It is useful for preparation of the data in the IHM setting, which consists
# of 8 channels, each belonging to a different speaker. This mapping provides the correspondence
# between the IHM channel and the speaker, which is then used in preparing the supervisions.
SPK_TO_CHANNEL_MAP = {
    "overlap_ratio_0.0_sil0.1_0.5_session0_actual0.0": {"1089": 5, "1320": 3, "1580": 0, "4077": 6, "4992": 1, "6829": 2, "6930": 7, "7176": 4},
    "overlap_ratio_0.0_sil0.1_0.5_session1_actual0.0": {"1089": 4, "121": 3, "2961": 0, "3575": 2, "5105": 6, "6829": 5, "8463": 7, "8555": 1},
    "overlap_ratio_0.0_sil0.1_0.5_session2_actual0.0": {"2961": 3, "4970": 0, "5105": 7, "5639": 5, "61": 2, "7176": 6, "7729": 4, "8224": 1},
    "overlap_ratio_0.0_sil0.1_0.5_session3_actual0.0": {"1089": 5, "1320": 4, "260": 3, "5105": 7, "61": 2, "672": 1, "6829": 6, "908": 0},
    "overlap_ratio_0.0_sil0.1_0.5_session4_actual0.0": {"1188": 2, "1221": 6, "1995": 1, "2961": 5, "4507": 7, "4970": 0, "5683": 3, "672": 4},
    "overlap_ratio_0.0_sil0.1_0.5_session5_actual0.0": {"121": 5, "1221": 4, "2300": 2, "237": 3, "4507": 6, "4970": 0, "7021": 1, "8463": 7},
    "overlap_ratio_0.0_sil0.1_0.5_session6_actual0.0": {"260": 5, "3575": 4, "3729": 2, "4507": 6, "4970": 0, "5683": 1, "6829": 7, "7729": 3},
    "overlap_ratio_0.0_sil0.1_0.5_session7_actual0.0": {"121": 4, "2300": 1, "260": 3, "3729": 2, "4077": 5, "8224": 0, "8230": 6, "8463": 7},
    "overlap_ratio_0.0_sil0.1_0.5_session8_actual0.0": {"1188": 2, "1995": 1, "237": 5, "3570": 0, "5639": 6, "5683": 3, "61": 4, "7127": 7},
    "overlap_ratio_0.0_sil0.1_0.5_session9_actual0.0": {"61": 2, "672": 3, "6930": 1, "7021": 0, "7127": 5, "7729": 4, "8230": 6, "8463": 7},
    "overlap_ratio_0.0_sil2.9_3.0_session0_actual0.0": {"121": 7, "260": 5, "3575": 0, "5105": 2, "5683": 1, "6930": 6, "8224": 4, "8230": 3},
    "overlap_ratio_0.0_sil2.9_3.0_session1_actual0.0": {"1284": 6, "1580": 5, "237": 7, "2961": 3, "3575": 0, "4446": 1, "4507": 4, "7127": 2},
    "overlap_ratio_0.0_sil2.9_3.0_session2_actual0.0": {"1188": 4, "121": 6, "1995": 7, "4446": 1, "7021": 5, "7729": 0, "8463": 3, "8555": 2},
    "overlap_ratio_0.0_sil2.9_3.0_session3_actual0.0": {"1995": 7, "2094": 3, "2830": 4, "2961": 2, "3729": 6, "4992": 1, "5105": 0, "7021": 5},
    "overlap_ratio_0.0_sil2.9_3.0_session4_actual0.0": {"1089": 6, "1188": 5, "2961": 2, "7021": 7, "7729": 0, "8230": 1, "8463": 4, "8555": 3},
    "overlap_ratio_0.0_sil2.9_3.0_session5_actual0.0": {"1580": 3, "2094": 2, "260": 6, "3729": 5, "4992": 0, "672": 1, "7021": 4, "8455": 7},
    "overlap_ratio_0.0_sil2.9_3.0_session6_actual0.0": {"1188": 4, "1320": 7, "1995": 6, "2300": 0, "3729": 5, "4507": 2, "7127": 1, "8455": 3},
    "overlap_ratio_0.0_sil2.9_3.0_session7_actual0.0": {"1089": 5, "1320": 7, "2830": 4, "4077": 3, "4992": 2, "7127": 1, "8230": 0, "908": 6},
    "overlap_ratio_0.0_sil2.9_3.0_session8_actual0.0": {"2961": 2, "4992": 1, "5142": 7, "672": 0, "6930": 6, "7176": 5, "8463": 3, "908": 4},
    "overlap_ratio_0.0_sil2.9_3.0_session9_actual0.0": {"1089": 5, "1188": 4, "2300": 0, "260": 7, "4077": 3, "672": 1, "8555": 2, "908": 6},
    "overlap_ratio_10.0_sil0.1_1.0_session0_actual10.1": {"1320": 6, "1995": 1, "260": 7, "4992": 0, "672": 4, "6930": 5, "8455": 2, "8463": 3},
    "overlap_ratio_10.0_sil0.1_1.0_session1_actual10.2": {"1188": 7, "1580": 1, "2094": 5, "3570": 3, "8224": 6, "8463": 4, "8555": 0, "908": 2},
    "overlap_ratio_10.0_sil0.1_1.0_session2_actual10.0": {"1188": 7, "3570": 3, "3729": 1, "5683": 4, "61": 6, "7127": 0, "7729": 5, "8463": 2},
    "overlap_ratio_10.0_sil0.1_1.0_session3_actual10.1": {"1580": 2, "1995": 3, "2300": 1, "3575": 7, "672": 4, "6829": 0, "7729": 6, "8224": 5},
    "overlap_ratio_10.0_sil0.1_1.0_session4_actual10.0": {"1188": 7, "121": 6, "2300": 1, "260": 5, "672": 2, "6829": 0, "7021": 3, "8224": 4},
    "overlap_ratio_10.0_sil0.1_1.0_session5_actual9.9": {"237": 0, "3575": 7, "3729": 3, "4507": 1, "4970": 6, "672": 4, "6930": 5, "8230": 2},
    "overlap_ratio_10.0_sil0.1_1.0_session6_actual9.9": {"1089": 5, "121": 3, "1320": 6, "3575": 7, "4446": 4, "4992": 2, "6829": 0, "8555": 1},
    "overlap_ratio_10.0_sil0.1_1.0_session7_actual10.1": {"121": 7, "1221": 5, "1995": 3, "4077": 0, "61": 1, "7729": 6, "8463": 4, "908": 2},
    "overlap_ratio_10.0_sil0.1_1.0_session8_actual10.0": {"1320": 5, "1580": 1, "2300": 0, "4077": 7, "4446": 3, "672": 4, "7176": 2, "7729": 6},
    "overlap_ratio_10.0_sil0.1_1.0_session9_actual10.0": {"1320": 5, "2830": 3, "3570": 7, "5639": 1, "6930": 4, "8224": 6, "8455": 2, "8555": 0},
    "overlap_ratio_20.0_sil0.1_1.0_session0_actual20.8": {"1089": 2, "121": 1, "1284": 0, "4507": 4, "4970": 7, "6930": 5, "7127": 6, "8555": 3},
    "overlap_ratio_20.0_sil0.1_1.0_session1_actual20.5": {"1089": 0, "1320": 6, "1580": 2, "260": 1, "4446": 4, "5105": 7, "5142": 3, "8224": 5},
    "overlap_ratio_20.0_sil0.1_1.0_session2_actual21.1": {"1580": 2, "2830": 6, "2961": 5, "3570": 7, "4507": 4, "5639": 1, "6829": 3, "8230": 0},
    "overlap_ratio_20.0_sil0.1_1.0_session3_actual20.0": {"1320": 5, "260": 1, "4992": 7, "5105": 6, "5142": 3, "7729": 2, "8455": 4, "908": 0},
    "overlap_ratio_20.0_sil0.1_1.0_session4_actual20.0": {"1089": 1, "1580": 2, "2830": 3, "3570": 7, "3729": 5, "5105": 4, "7127": 6, "8230": 0},
    "overlap_ratio_20.0_sil0.1_1.0_session5_actual19.6": {"1089": 1, "1188": 5, "1284": 0, "2961": 3, "3570": 7, "3575": 6, "61": 2, "8455": 4},
    "overlap_ratio_20.0_sil0.1_1.0_session6_actual20.0": {"121": 0, "4446": 5, "4507": 4, "5105": 7, "6829": 2, "7176": 6, "8224": 3, "8463": 1},
    "overlap_ratio_20.0_sil0.1_1.0_session7_actual20.1": {"2300": 0, "237": 5, "2830": 2, "2961": 1, "4970": 7, "4992": 6, "672": 4, "6930": 3},
    "overlap_ratio_20.0_sil0.1_1.0_session8_actual19.8": {"1221": 2, "1995": 4, "2300": 1, "672": 6, "7127": 7, "8224": 5, "8230": 0, "908": 3},
    "overlap_ratio_20.0_sil0.1_1.0_session9_actual20.7": {"1089": 2, "1284": 0, "4077": 3, "4446": 5, "5105": 7, "5639": 1, "7176": 6, "7729": 4},
    "overlap_ratio_30.0_sil0.1_1.0_session0_actual29.7": {"1089": 1, "1995": 6, "237": 7, "2830": 0, "2961": 4, "3575": 2, "672": 3, "7021": 5},
    "overlap_ratio_30.0_sil0.1_1.0_session1_actual30.4": {"1580": 6, "3575": 4, "4970": 2, "4992": 1, "5142": 5, "7729": 0, "8230": 7, "8455": 3},
    "overlap_ratio_30.0_sil0.1_1.0_session2_actual29.6": {"1284": 6, "1995": 7, "3575": 4, "4507": 2, "5639": 3, "61": 5, "8224": 0, "8463": 1},
    "overlap_ratio_30.0_sil0.1_1.0_session3_actual30.2": {"1320": 7, "2094": 3, "260": 4, "3575": 5, "4446": 6, "5105": 0, "6930": 1, "7729": 2},
    "overlap_ratio_30.0_sil0.1_1.0_session4_actual29.8": {"121": 1, "1320": 7, "260": 6, "2830": 5, "5683": 4, "6829": 0, "8463": 3, "8555": 2},
    "overlap_ratio_30.0_sil0.1_1.0_session5_actual29.7": {"1089": 5, "260": 4, "2830": 3, "3729": 2, "4077": 1, "4446": 6, "8224": 0, "908": 7},
    "overlap_ratio_30.0_sil0.1_1.0_session6_actual30.1": {"2094": 2, "237": 6, "4992": 4, "5683": 3, "61": 5, "6829": 1, "8555": 0, "908": 7},
    "overlap_ratio_30.0_sil0.1_1.0_session7_actual30.2": {"1089": 5, "1188": 1, "1284": 6, "2300": 4, "2830": 2, "3570": 0, "4446": 7, "4970": 3},
    "overlap_ratio_30.0_sil0.1_1.0_session8_actual29.7": {"1188": 1, "1284": 4, "3570": 0, "3575": 3, "4970": 2, "7021": 5, "8230": 6, "908": 7},
    "overlap_ratio_30.0_sil0.1_1.0_session9_actual29.8": {"1188": 2, "1320": 7, "61": 3, "6930": 1, "7021": 4, "7127": 5, "7176": 6, "7729": 0},
    "overlap_ratio_40.0_sil0.1_1.0_session0_actual39.5": {"121": 0, "1284": 7, "1320": 1, "2830": 5, "3729": 6, "4446": 3, "7127": 2, "7729": 4},
    "overlap_ratio_40.0_sil0.1_1.0_session1_actual39.7": {"121": 0, "1580": 2, "237": 3, "260": 1, "4446": 4, "7021": 7, "7729": 5, "8455": 6},
    "overlap_ratio_40.0_sil0.1_1.0_session2_actual41.2": {"1188": 3, "1284": 7, "1320": 4, "260": 1, "4507": 0, "6930": 5, "8224": 6, "8230": 2},
    "overlap_ratio_40.0_sil0.1_1.0_session3_actual40.2": {"1320": 4, "1580": 1, "3575": 2, "4077": 5, "4970": 0, "5105": 7, "7127": 6, "8463": 3},
    "overlap_ratio_40.0_sil0.1_1.0_session4_actual39.0": {"1188": 1, "121": 0, "1995": 2, "3729": 7, "4077": 3, "7729": 5, "8555": 4, "908": 6},
    "overlap_ratio_40.0_sil0.1_1.0_session5_actual42.0": {"1089": 4, "1284": 7, "237": 2, "2961": 6, "4077": 1, "4446": 3, "4507": 0, "8224": 5},
    "overlap_ratio_40.0_sil0.1_1.0_session6_actual39.9": {"1188": 1, "2094": 4, "3575": 5, "4970": 0, "5105": 7, "672": 2, "7021": 6, "8230": 3},
    "overlap_ratio_40.0_sil0.1_1.0_session7_actual40.5": {"1221": 2, "1580": 1, "2830": 6, "5142": 4, "7021": 7, "8230": 0, "8455": 5, "8463": 3},
    "overlap_ratio_40.0_sil0.1_1.0_session8_actual40.5": {"1580": 4, "260": 3, "3729": 7, "4970": 2, "5639": 6, "61": 0, "6930": 5, "8230": 1},
    "overlap_ratio_40.0_sil0.1_1.0_session9_actual39.9": {"1284": 7, "1995": 0, "2961": 6, "3575": 2, "4077": 4, "7176": 1, "8224": 5, "8463": 3}
}

OVERLAP_RATIOS = ["0L", "0S", "OV10", "OV20", "OV30", "OV40"]
# fmt: on


def download_libricss(target_dir: Pathlike, force_download: bool = False) -> Path:
    """
    Downloads the LibriCSS data from the Google Drive and extracts it.
    :param target_dir: the directory where the LibriCSS data will be saved.
    :param force_download: if True, it will download the LibriCSS data even if it is already present.
    :return: the path to downloaded and extracted directory with data.
    """
    # Download command (taken from https://github.com/chenzhuo1011/libri_css/blob/9e3b7b0c9bffd8ef6da19f7056f3a2f2c2484ffa/dataprep/scripts/dataprep.sh#L27)
    command = """wget --load-cookies /tmp/cookies.txt "https://docs.google.com/uc?export=download&confirm=$(wget --quiet --save-cookies /tmp/cookies.txt --keep-session-cookies --no-check-certificate 'https://docs.google.com/uc?export=download&id=1Piioxd5G_85K9Bhcr8ebdhXx0CnaHy7l' -O- | sed -rn 's/.*confirm=([0-9A-Za-z_]+).*/\1\n/p')&id=1Piioxd5G_85K9Bhcr8ebdhXx0CnaHy7l" -O for_release.zip && rm -rf /tmp/cookies.txt"""

    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    corpus_zip = target_dir / "for_release.zip"
    corpus_dir = target_dir / "for_release"

    if not force_download and corpus_zip.exists():
        logging.info(f"{corpus_zip} already exists. Skipping download.")
    else:
        subprocess.run(command, shell=True, cwd=target_dir)

    # Extract the zipped file
    if not corpus_dir.exists() or force_download:
        logging.info(f"Extracting {corpus_zip} to {target_dir}")
        with zipfile.ZipFile(corpus_zip, "r") as corpus_zip:
            for member in tqdm(corpus_zip.infolist(), desc="Extracting"):
                corpus_zip.extract(member, target_dir)

    return target_dir


def prepare_libricss(
    corpus_dir: Pathlike,
    output_dir: Pathlike = None,
    type: str = "mdm",
    segmented_cuts: bool = False,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet, CutSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    NOTE: The recordings contain all 7 channels. If you want to use only one channel, you can
    use either ``recording.load_audio(channel=0)`` or ``MonoCut(id=...,recording=recording,channel=0)``
    while creating the CutSet.

    :param corpus_dir: Pathlike, the path to the extracted corpus.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param type: str, the type of data to prepare ('mdm', 'sdm', 'ihm-mix', or 'ihm'). These settings
        are similar to the ones in AMI and ICSI recipes.
    :param segmented_cuts: bool, if True, it will return 1-minute (as described in the original paper)
        in the form of a CutSet. These are saved under the index ``segments`` in the returned Dict.
        May be useful for evaluating multi-talker ASR systems, e.g., in this paper: https://arxiv.org/abs/2109.08555.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.

    """
    assert type in ["mdm", "sdm", "ihm-mix", "ihm"]

    corpus_dir = Path(corpus_dir)
    corpus_dir = (
        corpus_dir / "for_release" if corpus_dir.stem != "for_release" else corpus_dir
    )

    recordings = []
    segments = []
    session_name_map = {}  # Map from original session name to simplified name

    for ov in OVERLAP_RATIOS:
        for session in (corpus_dir / ov).iterdir():
            _, _, _, _, _, name, actual_ov = session.name.split("_")
            actual_ov = float(actual_ov.split("actual")[1])
            recording_id = f"{ov}_{name}"
            session_name_map[session.name] = recording_id
            audio_path = (
                session / "clean" / "mix.wav"
                if type == "ihm-mix"
                else session / "clean" / "each_spk.wav"
                if type == "ihm"
                else session / "record" / "raw_recording.wav"
            )
            recording = Recording.from_file(audio_path, recording_id=recording_id)

            if type == "sdm":
                recordings.append(fastcopy(recording, channel_ids=[0]))
            else:
                recordings.append(recording)

            for idx, seg in enumerate(
                parse_transcript(session / "transcription" / "meeting_info.txt")
            ):
                if type == "ihm-mix" or type == "sdm":
                    channel = 0
                elif type == "ihm":
                    channel = SPK_TO_CHANNEL_MAP[session.name][seg[2]]
                else:
                    channel = list(range(7))
                segments.append(
                    SupervisionSegment(
                        id=f"{recording_id}-{idx}",
                        recording_id=recording_id,
                        start=seg[0],
                        duration=seg[1] - seg[0],
                        text=seg[4],
                        language="English",
                        speaker=seg[2],
                        channel=channel,
                    )
                )

    supervisions = SupervisionSet.from_segments(segments)
    recordings = RecordingSet.from_recordings(recordings)
    recordings, supervisions = fix_manifests(recordings, supervisions)
    validate_recordings_and_supervisions(recordings, supervisions)

    result_dict = {"recordings": recordings, "supervisions": supervisions}

    if segmented_cuts:
        segments = defaultdict(list)
        # Read the segments from the JSON file. The JSON file is a dictionary
        # with the session name as the key and the value is a list of tuples
        # with the start and end time of the segments.
        with open(corpus_dir / "all_res.json", "r") as f:
            res = json.load(f)
            for session, segs in res.items():
                segments[session_name_map[session]].extend(segs)

        # Create recording-level cuts. We will then truncate them to create
        # the 1-minute segments.
        cuts_reco = CutSet.from_manifests(
            recordings=recordings,
            supervisions=supervisions,
        )

        # Now create the 1-minute segments.
        cuts_segmented = []
        for session, session_segments in segments.items():
            # Get the recording-level cut
            session_cut = cuts_reco.filter(
                lambda c: c.recording_id == session
            ).to_eager()[0]

            # Iterate over all the segments and create a new cut for each
            for idx, seg in enumerate(session_segments):
                start = seg[0] / 16000
                duration = (seg[1] - seg[0]) / 16000
                new_cut = session_cut.truncate(
                    offset=start,
                    duration=duration,
                    keep_excessive_supervisions=False,
                ).with_id(f"{session}-{idx}")
                cuts_segmented.append(new_cut)

        # Add the segmented cuts to the result dict
        cuts_segmented = CutSet.from_cuts(cuts_segmented)
        result_dict["segments"] = cuts_segmented

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        recordings.to_file(output_dir / f"libricss-{type}_recordings_all.jsonl.gz")
        supervisions.to_file(output_dir / f"libricss-{type}_supervisions_all.jsonl.gz")
        if segmented_cuts:
            cuts_segmented.to_file(
                output_dir / f"libricss-{type}_segments_all.jsonl.gz"
            )

    return result_dict


def parse_transcript(file_name):
    """
    Parses the transcript file and returns a list of SupervisionSegment objects.
    """
    segments = []
    with open(file_name, "r") as f:
        next(f)  # skip the first line
        for line in f:
            start, end, speaker, utt_id, text = line.split("\t")
            segments.append((float(start), float(end), speaker, utt_id, text))
    return segments



"""
╔═════════════════════════════════════════════╗
║ Export/Import CutSet to HuggingFace Dataset ║
╚═════════════════════════════════════════════╝
"""
from hashlib import md5
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from lhotse import Image, Recording, SupervisionSegment
from lhotse.cut import CutSet, MonoCut
from lhotse.utils import Pathlike, is_module_available


def contains_only_mono_cuts(cutset: CutSet) -> bool:
    return all(isinstance(cut, MonoCut) for cut in cutset)


def has_one_supervision_per_cut(cutset: CutSet) -> bool:
    return all(len(cut.supervisions) == 1 for cut in cutset)


def has_one_audio_source(cutset: CutSet) -> bool:
    return all(len(cut.recording.sources) == 1 for cut in cutset)


def convert_cuts_info_to_hf(cutset: CutSet) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Converts the cut information into a dictionary compatible with HuggingFace datasets format.

    :param cutset: A CutSet object.
    :return: A tuple where the first element is a dictionary
        representing the cut attributes and the second element is a dictionary describing the
        format of the HuggingFace dataset.
    """
    from datasets import Audio, Value

    cut_info = {
        "id": [cut.id for cut in cutset],
        "audio": [cut.recording.sources[0].source for cut in cutset],
        "duration": [cut.duration for cut in cutset],
        "num_channels": [len(cut.recording.channel_ids) for cut in cutset],
    }
    cut_info_description = {
        "id": Value("string"),
        "audio": Audio(mono=False),
        "duration": Value("float"),
        "num_channels": Value("uint16"),
    }
    return cut_info, cut_info_description


def convert_supervisions_info_to_hf(
    cutset: CutSet,
    exclude_attributes: Optional[Union[List[str], Set[str]]] = None,
) -> Tuple[List[List[Dict[str, Any]]], Dict[str, Any]]:
    """
    Converts cut supervisions into a dictionary compatible with HuggingFace datasets format.

    :param cutset: A CutSet object.
    :param exclude_attributes: A list|set of attributes to exclude from the supervisions dicts.
    :return: A tuple where the first element is a dictionary
        representing the cut attributes and the second element is a dictionary describing the
        format of the HuggingFace dataset.
    """

    from datasets import Features, Sequence, Value

    has_speaker = any(
        (
            hasattr(cut.supervisions[0], "speaker")
            and cut.supervisions[0].speaker is not None
        )
        for cut in cutset
    )
    has_language = any(
        (
            hasattr(cut.supervisions[0], "language")
            and cut.supervisions[0].language is not None
        )
        for cut in cutset
    )
    alignment_types = [
        s.alignment.keys()
        for c in cutset
        for s in c.supervisions
        if s.alignment is not None
    ]
    alignment_types = set([item for sublist in alignment_types for item in sublist])

    sup_dicts = []
    for c in cutset:
        cut_sup_dicts = []
        for s in c.supervisions:
            sup_dict = {
                "text": s.text,
            }

            if exclude_attributes is None or "start" not in exclude_attributes:
                sup_dict["start"] = s.start

            if exclude_attributes is None or "end" not in exclude_attributes:
                sup_dict["end"] = s.end

            if exclude_attributes is None or "channel" not in exclude_attributes:
                if isinstance(s.channel, list):
                    sup_dict["channel"] = ",".join(map(str, s.channel))
                else:
                    sup_dict["channel"] = str(s.channel)

            if has_speaker and (
                exclude_attributes is None or "speaker" not in exclude_attributes
            ):
                sup_dict["speaker"] = str(s.speaker)

            if has_language and (
                exclude_attributes is None or "language" not in exclude_attributes
            ):
                sup_dict["language"] = str(s.language)

            if alignment_types and (
                exclude_attributes is None or "alignments" not in exclude_attributes
            ):
                alignments = {}
                for alignment_type in alignment_types:
                    alignments[alignment_type + "_alignment"] = list(
                        map(
                            lambda item: {
                                "symbol": item.symbol,
                                "start": item.start,
                                "end": item.end,
                            },
                            s.alignment[alignment_type],
                        )
                    )

                sup_dict = {**sup_dict, **alignments}

            cut_sup_dicts.append(sup_dict)
        sup_dicts.append(cut_sup_dicts)

    sup_dicts_info = {"text": Value("string")}

    if exclude_attributes is None or "start" not in exclude_attributes:
        sup_dicts_info["start"] = Value("float")

    if exclude_attributes is None or "end" not in exclude_attributes:
        sup_dicts_info["end"] = Value("float")

    if exclude_attributes is None or "channel" not in exclude_attributes:
        sup_dicts_info["channel"] = Value("string")

    if has_speaker and (
        exclude_attributes is None or "speaker" not in exclude_attributes
    ):
        sup_dicts_info["speaker"] = Value("string")

    if has_language and (
        exclude_attributes is None or "language" not in exclude_attributes
    ):
        sup_dicts_info["language"] = Value("string")

    if alignment_types and (
        exclude_attributes is None or "alignments" not in exclude_attributes
    ):
        alignment_info = {
            "symbol": Value("string"),
            "start": Value("float"),
            "end": Value("float"),
        }
        for alignment_type in alignment_types:
            sup_dicts_info[alignment_type + "_alignment"] = Sequence(
                Features(**alignment_info)
            )

    return sup_dicts, sup_dicts_info


def lod_to_dol(lod: List[Dict[str, Any]]) -> Dict[str, List]:
    """
    Converts List of Dicts to Dict of Lists.
    """
    return {k: [d[k] for d in lod] for k in lod[0].keys()}


def export_cuts_to_hf(cutset: CutSet):
    """
    Converts a CutSet to a HuggingFace Dataset. Currently, only MonoCut with one recording source is supported.
    Other cut types will be supported in the future.

    Currently, two formats are supported:
        1. If each cut has one supervision (e.g. LibriSpeech), each cut is represented as a single row (entry)
           in the HuggingFace dataset with all the supervision information stored along the cut information.
           The final HuggingFace dataset format is:
               ╔═══════════════════╦═══════════════════════════════╗
               ║      Feature      ║            Type               ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║        id         ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║      audio        ║ Audio()                       ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║     duration      ║ Value(dtype='float32')        ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║   num_channels    ║ Value(dtype='uint16')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║       text        ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║     speaker       ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║     language      ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║   {x}_alignment   ║ Sequence(Alignment)           ║
               ╚═══════════════════╩═══════════════════════════════╝
           where x stands for the alignment type (commonly used: "word", "phoneme").

           Alignment is represented as:
               ╔═══════════════════╦═══════════════════════════════╗
               ║      Feature      ║            Type               ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║      symbol       ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║       start       ║ Value(dtype='float32')        ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║        end        ║ Value(dtype='float32')        ║
               ╚═══════════════════╩═══════════════════════════════╝


        2. If each cut has multiple supervisions (e.g. AMI), each cut is represented as a single row (entry)
           while all the supervisions are stored in a separate list of dictionaries under the 'segments' key.
           The final HuggingFace dataset format is:
               ╔══════════════╦════════════════════════════════════╗
               ║   Feature    ║                 Type               ║
               ╠══════════════╬════════════════════════════════════╣
               ║      id      ║ Value(dtype='string')              ║
               ╠══════════════╬════════════════════════════════════╣
               ║    audio     ║ Audio()                            ║
               ╠══════════════╬════════════════════════════════════╣
               ║   duration   ║ Value(dtype='float32')             ║
               ╠══════════════╬════════════════════════════════════╣
               ║ num_channels ║ Value(dtype='uint16')              ║
               ╠══════════════╬════════════════════════════════════╣
               ║   segments   ║ Sequence(Segment)                  ║
               ╚══════════════╩════════════════════════════════════╝
           where one Segment is represented as:
               ╔═══════════════════╦═══════════════════════════════╗
               ║      Feature      ║            Type               ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║        text       ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║       start       ║ Value(dtype='float32')        ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║        end        ║ Value(dtype='float32')        ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║      channel      ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║      speaker      ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║      language     ║ Value(dtype='string')         ║
               ╠═══════════════════╬═══════════════════════════════╣
               ║   {x}_alignment   ║ Sequence(Alignment)           ║
               ╚═══════════════════╩═══════════════════════════════╝

    :param cutset: A CutSet object.
    :return: A HuggingFace Dataset.
    """

    assert has_one_audio_source(
        cutset
    ), "Only CutSets with one audio source per cut are supported. MultiSource cuts coming soon."

    if not is_module_available("datasets"):
        raise ImportError(
            "Please install the 'datasets' package (pip install datasets)."
        )
    from datasets import Dataset, Features, Sequence

    # We don't need start and end attribute if we have only one supervision/segment per cut,
    #  as start=0 and end=duration.
    cut_info, cut_info_description = convert_cuts_info_to_hf(cutset)
    sup_dicts, sup_dicts_info = convert_supervisions_info_to_hf(
        cutset,
        exclude_attributes={"start", "end", "channel"}
        if has_one_supervision_per_cut(cutset)
        else None,
    )

    if has_one_supervision_per_cut(cutset):
        dataset_dict = {
            **cut_info,
            **lod_to_dol([x[0] for x in sup_dicts]),
        }
        dataset_info = Features(
            **cut_info_description,
            **sup_dicts_info,
        )
    else:
        dataset_dict = {
            **cut_info,
            "segments": sup_dicts,
        }
        dataset_info = Features(
            segments=Sequence(Features(**sup_dicts_info)),
            **cut_info_description,
        )

    return Dataset.from_dict(dataset_dict, features=dataset_info)


class LazyHFDatasetIterator:
    """
    Thin wrapper on top of HF datasets objects that allows to interact with them through a Lhotse CutSet.
    It can be initialized with an existing HF dataset, or args/kwargs passed on to ``datasets.load_dataset()``.

    Use ``audio_key``, ``text_key``, ``lang_key`` and ``gender_key`` options to indicate which keys in dict examples
    returned from HF Dataset should be looked up for audio, transcript, language, and gender respectively.
    The remaining keys in HF dataset examples will be stored inside ``cut.custom`` dictionary.

    Example with existing HF dataset::

        >>> import datasets
        ... dataset = datasets.load_dataset("mozilla-foundation/common_voice_11_0", "hi", split="test")
        ... dataset = dataset.map(some_transform)
        ... cuts_it = LazyHFDatasetIterator(dataset)
        ... for cut in cuts_it:
        ...     pass

    Example providing HF dataset init args/kwargs::

        >>> import datasets
        ... cuts_it = LazyHFDatasetIterator("mozilla-foundation/common_voice_11_0", "hi", split="test")
        ... for cut in cuts_it:
        ...     pass

    """

    def __init__(
        self,
        *dataset_args,
        audio_key: str = "audio",
        text_key: str = "sentence",
        lang_key: str = "language",
        gender_key: str = "gender",
        **dataset_kwargs
    ):
        assert is_module_available("datasets")
        self.audio_key = audio_key
        self.text_key = text_key
        self.lang_key = lang_key
        self.gender_key = gender_key
        self.dataset_args = dataset_args
        self.dataset_kwargs = dataset_kwargs

    def __iter__(self):
        from datasets import (
            Audio,
            Dataset,
            DatasetDict,
            IterableDataset,
            IterableDatasetDict,
            load_dataset,
        )

        if len(self.dataset_args) == 1 and isinstance(
            self.dataset_args[0],
            (Dataset, IterableDataset, DatasetDict, IterableDatasetDict),
        ):
            dataset = self.dataset_args[0]
        else:
            dataset = load_dataset(*self.dataset_args, **self.dataset_kwargs)
        dataset = dataset.cast_column(self.audio_key, Audio(decode=False))
        for item in dataset:
            audio_data = item.pop(self.audio_key)
            recording = Recording.from_bytes(
                audio_data["bytes"], recording_id=md5(audio_data["bytes"]).hexdigest()
            )
            supervision = SupervisionSegment(
                id=recording.id,
                recording_id=recording.id,
                start=0.0,
                duration=recording.duration,
                text=item.pop(self.text_key, None),
                language=item.pop(self.lang_key, None),
                gender=item.pop(self.gender_key, None),
            )
            cut = recording.to_cut()
            cut.supervisions = [supervision]
            maybe_resolve_images(item)
            cut.custom = item
            yield cut


def maybe_resolve_images(item: dict) -> None:
    if not is_module_available("PIL"):
        return
    import PIL.Image

    for k, v in item.items():
        if isinstance(v, PIL.Image.Image):
            item[k] = Image.from_pillow(v)



from .image import Image
from .io import (
    PillowInMemoryReader,
    PillowInMemoryWriter,
    PillowReader,
    PillowWriter,
    available_storage_backends,
    get_reader,
    get_writer,
    register_reader,
    register_writer,
)



"""
This is an experimental workflow that can be used to simulate multi-speaker meetings from
a CutSet containing MonoCut objects.
"""
import abc
import random
from itertools import groupby
from typing import List, Optional, Union

import numpy as np
from tqdm import tqdm

from lhotse import RecordingSet, SupervisionSet
from lhotse.cut import CutSet
from lhotse.dataset.sampling import DynamicCutSampler
from lhotse.utils import fastcopy, is_module_available

MAX_TASKS_WAITING = 1000


class BaseMeetingSimulator(abc.ABC):
    """
    Base class for meeting simulators. A simulator consists of a `fit()`, a `simulate()`,
    and a `reverberate()` method.

    The `fit()` method is used to learn the distribution of the meeting parameters
    (e.g. turn-taking, overlap ratio, etc.) from a given dataset, presented in the form of
    a SupervisionSet. The parameters themselves are simulator specific.

    The `simulate()` method takes a CutSet containing MonoCut objects and simulates the
    desired number of multi-speaker meetings, based on the learned distribution. The output
    is a CutSet containing MixedCut objects, where each track represents a different speaker.

    The `reverberate()` method takes a CutSet containing MixedCut objects (usually the output
    of `simulate()`) and applies a reverberation effect to each track. We can apply single
    or multi-channel room impulse responses (RIRs) to each track. The output is a CutSet
    containing MixedCut objects, where each track represents a different speaker, convolved
    with a different RIR.

    The base class should be inherited from and the different methods should be implemented.

    The output is expected to be a CutSet containing MixedCut objects, where each track
    represents a different speaker, possibly convolved with a different RIR. This is
    analogous to the "mixture model" of speech signals.

    Example usage:
    >>> simulator = MyMeetingSimulator()
    >>> simulator.fit(cuts)
    >>> simulated_cuts = simulator.simulate(mono_cuts, num_meetings=10)
    >>> simulated_cuts = simulator.reverberate(simulated_cuts, rirs)
    """

    def __init__(self):
        if type(self) is BaseMeetingSimulator:
            raise TypeError(
                "BaseMeetingSimulator is an abstract base class and should not be instantiated."
            )

        if not is_module_available("scipy"):
            raise ImportError("Please 'pip install scipy' first.")

    def __repr__(self) -> str:
        return f"{type(self).__name__}()"

    @abc.abstractmethod
    def fit(self, meetings: Optional[SupervisionSet] = None) -> None:
        """
        Learn the distribution of the meeting parameters from a given dataset.
        """
        ...

    @abc.abstractmethod
    def simulate(
        self,
        cuts: CutSet,
        num_meetings: Optional[int] = None,
        num_repeats: Optional[int] = None,
    ) -> CutSet:
        """
        Simulate the desired number of multi-speaker meetings.
        """
        ...

    @abc.abstractmethod
    def reverberate(self, cuts: CutSet, *rirs: RecordingSet) -> CutSet:
        """
        Apply a reverberation effect to each track.
        """
        ...


class MeetingSampler:
    """
    Create a sampler that will be used to sample groups of utterances from the sources.
    The cuts are partitioned into speaker-wise buckets, and a SimpleCutSampler is created
    for each bucket. When we sample a group of utterances, we first sample the number of
    speakers in the meeting, and then sample the utterances of each speaker. This is done
    by sampling a batch from the corresponding SimpleCutSampler.

    :param cuts: a CutSet containing MonoCut objects.
    :param num_repeats: the number of times each cut will be repeated (by default, they
        are repeated infinitely).
    :param num_meetings: the number of meetings to simulate.
    :param num_speakers_per_meeting: the number of speakers per meeting.
    :param speaker_count_probs: the probabilities of the number of speakers per meeting.
    :param max_duration_per_speaker: the maximum duration of a speaker in a meeting.
    :param max_utterances_per_speaker: the maximum number of utterances of a speaker in a meeting.
    :param seed: the random seed.
    :return: a DynamicCutSampler object.
    """

    def __init__(
        self,
        cuts: CutSet,
        num_repeats: Optional[int] = None,
        num_meetings: Optional[int] = None,
        num_speakers_per_meeting: Union[int, List[int]] = 2,
        speaker_count_probs: Optional[List[float]] = None,
        max_duration_per_speaker: Optional[float] = 20.0,
        max_utterances_per_speaker: Optional[int] = 5,
        seed: int = 0,
    ):
        # Some basic checks
        assert all(n > 1 for n in num_speakers_per_meeting), (
            "The number of speakers per meeting must be greater than 1. "
            f"Got: {num_speakers_per_meeting}"
        )
        assert all(p > 0.0 for p in speaker_count_probs), (
            "The probabilities of the number of speakers per meeting must be greater than 0. "
            f"Got: {speaker_count_probs}"
        )
        assert sum(speaker_count_probs) == 1.0, (
            "The probabilities of the number of speakers per meeting must sum to 1. "
            f"Got: {speaker_count_probs}"
        )
        assert len(num_speakers_per_meeting) == len(
            speaker_count_probs
        ), "The number of speakers per meeting and the number of probabilities must be the same."

        # Create samplers for each bucket. We create this as a dict so that we can
        # efficiently remove items and also randomly sample items in constant time.
        # It also supports the len() function in constant time.
        # Note that a Python list is not a good choice here, because removing items
        # from a list is O(n). A set is also not a good choice, because randomly
        # sampling items from a set is O(n).
        self.samplers = {}
        for spk, spk_cuts in tqdm(
            groupby(
                sorted(cuts, key=lambda cut: cut.supervisions[0].speaker),
                lambda cut: cut.supervisions[0].speaker,
            ),
            desc="Creating samplers for each speaker...",
        ):
            sampler = DynamicCutSampler(
                CutSet.from_cuts(list(spk_cuts)).repeat(
                    times=num_repeats, preserve_id=False
                ),
                max_duration=max_duration_per_speaker,
                max_cuts=max_utterances_per_speaker,
                shuffle=True,
                seed=seed,
            )
            self.samplers[spk] = sampler

        self.num_speakers_per_meeting = num_speakers_per_meeting
        self.speaker_count_probs = speaker_count_probs

        self.npr = np.random.RandomState(seed)
        self.rng = random.Random(seed)
        self._remaining_meetings = num_meetings

    def __iter__(self):
        for sampler in self.samplers.values():
            iter(sampler)
        return self

    def __next__(self):
        # If we have sampled enough meetings, stop.
        if self._remaining_meetings is not None and self._remaining_meetings == 0:
            raise StopIteration()

        # If we don't have enough speakers, stop.
        if len(self.samplers) < min(self.num_speakers_per_meeting):
            raise StopIteration()

        # Sample the number of speakers for this meeting.
        N = min(
            self.npr.choice(self.num_speakers_per_meeting, p=self.speaker_count_probs),
            len(self.samplers),
        )

        # Sample speakers.
        this_batch_spk_ids = self.rng.sample(list(self.samplers.keys()), N)
        utterances = CutSet.from_cuts([])
        for spk_id in this_batch_spk_ids:
            sampler = self.samplers[spk_id]
            try:
                this_batch = next(sampler)
                utterances += this_batch
            except StopIteration:
                del self.samplers[spk_id]
                continue

        # shuffle the utterances
        utterances = utterances.shuffle()

        if self._remaining_meetings is not None:
            self._remaining_meetings -= 1
        return utterances if len(utterances) > 0 else next(self)


def reverberate_cuts(cuts: CutSet, *rirs: RecordingSet) -> CutSet:
    """
    Use provided RIRs to convolve each track of the input CutSet. The cuts here are
    MixedCut objects containing different speakers in different tracks. To reverberate,
    we choose a random RIR containing as many Recording objects as there are tracks
    in the MixedCut.

    If impulse responses are not provided, we use the fast randomized approximation
    method to simulate artificial single-channel RIRs.

    :param cuts: a CutSet containing MixedCut objects.
    :param rirs: one or more RecordingSet (each set is a group of RIRs from the same room).
    :return: a CutSet containing MixedCut objects reverberated with the provided RIRs.
    """
    out_cuts = []
    max_sources = max(len(rir_group) for rir_group in rirs) if len(rirs) > 0 else 0
    for cut in cuts:
        num_speakers = len(cut.tracks)
        if num_speakers <= max_sources:
            tracks = []
            # Choose a random RIR group containing as many recordings as there are
            # speakers (sources) in the mixed cut.
            rir_group = rirs.filter(lambda r: len(r) == num_speakers).random()
            for track, rir in zip(cut.tracks, rir_group):
                tracks.append(fastcopy(track, cut=track.cut.reverb_rir(rir)))
            out_cuts.append(fastcopy(cut, tracks=tracks))
        else:
            # We will use a fast random approximation to generate RIRs.
            out_cuts.append(cut.reverb_rir())

    return CutSet.from_cuts(out_cuts)
